/*
 *  File Name       : linux/arch/arm/mach-mp200/extio.c
 *  Function        : External GPIO
 *  Release Version : Ver 1.00
 *  Release Date    : 2009/06/11
 *
 *  Copyright (C) NEC Electronics Corporation 2009
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/i2c.h>

#include <asm/gpio.h>
#include <mach/extio.h>

#define EXTIO_OUT_INITDATA	0xff

static struct mutex	extio_lock;

int mp200_extio_initialized = 0;

static struct i2c_client *mp200_extio_client_in = NULL;
static struct i2c_client *mp200_extio_client_out = NULL;
static unsigned char extio_out_data = 0;


int extio_reg_read(int addr, unsigned char *data)
{
	int ret;

	if (mp200_extio_initialized == 0) {
		return -EBUSY;
	}

	if (addr == EXTIO_OUT) {
		mutex_lock(&extio_lock);
		ret = extio_out_data;
		mutex_unlock(&extio_lock);
	} else {
		ret = i2c_smbus_read_byte(mp200_extio_client_in);
		if (ret < 0) {
			printk("%s: failed reading\n", __FUNCTION__);
			return ret;
		}
	}
	*data = (unsigned char)ret;
	return 0;
}

int extio_reg_write(int addr, unsigned char data)
{
	int ret;

	if (mp200_extio_initialized == 0) {
		return -EBUSY;
	}

	if (addr == EXTIO_OUT) {
		mutex_lock(&extio_lock);
		ret = i2c_smbus_write_byte(mp200_extio_client_out, data);
		if (ret < 0) {
			printk("%s: failed writing\n", __FUNCTION__);
			mutex_unlock(&extio_lock);
			return ret;
		} else {
			extio_out_data = data;
		}
		mutex_unlock(&extio_lock);
	} else {
		return -EINVAL;
	}
	return 0;
}

int extio_read(int addr, unsigned char *data)
{
	return extio_reg_read(addr, data);
}

int extio_write(int addr, unsigned char data, unsigned char mask)
{
	int ret;
	unsigned char tmp_data;

	if (mp200_extio_initialized == 0) {
		return -EBUSY;
	}

	if (addr == EXTIO_OUT) {
		mutex_lock(&extio_lock);
		tmp_data = extio_out_data;
		tmp_data = (tmp_data & (~mask)) | (data & mask);
		ret = i2c_smbus_write_byte(mp200_extio_client_out, tmp_data);
		if (ret < 0) {
			printk("%s: failed writing\n", __FUNCTION__);
			mutex_unlock(&extio_lock);
			return ret;
		} else {
			extio_out_data = tmp_data;
		}
		mutex_unlock(&extio_lock);
	} else {
		return -EINVAL;
	}
	return 0;
}

int extio_set_direction(unsigned gpio, int is_input)
{
	if (is_input) {
		if (gpio >= GPIO_EXT2_P0 && gpio <= GPIO_EXT2_P7) {
			return 0;
		}
	} else {
		if (gpio >= GPIO_EXT2_P8 && gpio <= GPIO_EXT2_P15) {
			return 0;
		}
	}
	return -EINVAL;
}

int extio_get_value(unsigned int gpio)
{
	unsigned char val;
	unsigned int off;
	int ret = 0;

	if (mp200_extio_initialized == 0) {
		return -EBUSY;
	}

	if (gpio >= GPIO_EXT2_P0 && gpio <= GPIO_EXT2_P7) {
		off = gpio - GPIO_EXT2_P0;
		ret = extio_reg_read(EXTIO_IN, &val);
	} else if (gpio >= GPIO_EXT2_P8 && gpio <= GPIO_EXT2_P15) {
		off = gpio - GPIO_EXT2_P8;
		val = extio_out_data;
	} else {
		return -EINVAL;
	}

	return ret ? ret : (val & (1u << off));
}

void extio_set_value(unsigned int gpio, int value)
{
	unsigned int off;
	unsigned char val, mask;
	int ret = 0;

	if (mp200_extio_initialized == 0) {
		return;
	}

	if (gpio >= GPIO_EXT2_P8 && gpio <= GPIO_EXT2_P15) {
		mutex_lock(&extio_lock);
		off = gpio - GPIO_EXT2_P8;
		mask = 1u << (off & 0x7);
		val = extio_out_data;
		val = (value) ? val | mask : val & ~mask;
		ret = i2c_smbus_write_byte(mp200_extio_client_out, val);
		if (ret == 0) {
			extio_out_data = val;
		}
		mutex_unlock(&extio_lock);
	}
}

static int mp200_extio_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	if ((client->addr & 0x70) == 0x60) {
		if (!i2c_check_functionality(client->adapter,
					     I2C_FUNC_SMBUS_READ_BYTE_DATA)) {
			return -EIO;
		}
		mp200_extio_client_in = client;
	} else {
		if (!i2c_check_functionality(client->adapter,
					     I2C_FUNC_SMBUS_WRITE_BYTE_DATA)) {
			return -EIO;
		}
		mp200_extio_client_out = client;
		extio_out_data = EXTIO_OUT_INITDATA;
		i2c_smbus_write_byte(mp200_extio_client_out, extio_out_data);
	}

	if (mp200_extio_client_in != NULL && mp200_extio_client_out != NULL) {
		mp200_extio_initialized = 1;
		mutex_init(&extio_lock);
	}

	return 0;
}

static int mp200_extio_remove(struct i2c_client *client)
{
	mp200_extio_initialized = 0;
	return 0;
}

static struct i2c_device_id mp200_extio_idtable[] = {
	{ "max7324_w", 0 },
	{ "max7324_r", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, mp200_extio_idtable);

static struct i2c_driver mp200_extio_driver = {
	.driver = {
		.name	= "mp200-extio",
		.owner	= THIS_MODULE,
	},
	.id_table	= mp200_extio_idtable,
	.probe		= mp200_extio_probe,
	.remove		= __devexit_p(mp200_extio_remove),
};

static int __init mp200_extio_init(void)
{
	return i2c_add_driver(&mp200_extio_driver);
}

device_initcall(mp200_extio_init);

EXPORT_SYMBOL(extio_reg_read);
EXPORT_SYMBOL(extio_reg_write);
EXPORT_SYMBOL(extio_set_direction);
EXPORT_SYMBOL(extio_get_value);
EXPORT_SYMBOL(extio_set_value);
