#include <linux/ftrace.h>
#include <linux/init.h>
#include <linux/uaccess.h>

#include <asm/cacheflush.h>
#include <asm/sgidefs.h>
#include <asm/inst.h>
#include <asm/ftrace.h>

/*
 * On O32 the gcc generated code sequence to invoke _mcount looks like
 *
 *	move	$1, $31			# save current return address
 *	subu	$sp, $sp, 8		# _mcount pops 2 words from  stack
 *	jal	_mcount
 *	 j	$31
 *
 * If the _mcount call is nop'ed out we still need to pop the stack frame.
 * So we use an addiu $sp, $sp, 8 instead.  NABI doesn't allocate frames in
 * such a nonsenical way so a normal nop does the magic.
 */
static const union mips_instruction ftrace_nop = {
#if _MIPS_SIM == _MIPS_SIM_ABI32
	.i_format.opcode = addiu_op,		/* addiu $sp, $sp, 8 */
	.i_format.rs = 29,
	.i_format.rt = 29,
	.i_format.simmediate = 8
#else
	.word = 0				/* normal nop */
#endif
};

unsigned char *ftrace_nop_replace(void)
{
	return (unsigned char *) ftrace_nop.byte;
}

static unsigned char *ftrace_call_replace(unsigned long ip, unsigned long addr)
{
	static union mips_instruction inst;

	if (unlikely(ip ^ addr) >> 26) {
		WARN_ON_ONCE(1);
		return NULL;
	}

	inst.j_format.opcode = jal_op;
	inst.j_format.target = addr >> 2;

	return (unsigned char *) &inst;
}

int
ftrace_modify_code(unsigned long ip, unsigned char *old_code,
                   unsigned char *new_code)
{
	unsigned int old = *(unsigned int *) old_code;
	unsigned int new = *(unsigned int *) new_code;
	unsigned int replaced;
	int err;

	err = __get_user(replaced, (unsigned int *) ip);
	if (unlikely(err))
		return 1;

	err = __put_user(new, (unsigned int *) ip);
	if (unlikely(err))
		return 1;

	if (unlikely(replaced != old && replaced != new))
		return 2;

	/*
	 * This is broken.  We need to flush the icache on all processors
	 * but can't because ftrace_modify_code is executed with interrupts
	 * disabled ...
	 */
	local_flush_icache_range(ip, ip + MCOUNT_INSN_SIZE);

	return 0;
}

int ftrace_make_nop(struct module *mod,
		    struct dyn_ftrace *rec, unsigned long addr)
{
	unsigned char *new, *old;
	unsigned long ip = rec->ip;

	old = ftrace_call_replace(ip, addr);
	new = ftrace_nop_replace();

	return ftrace_modify_code(rec->ip, old, new);
}

int ftrace_make_call(struct dyn_ftrace *rec, unsigned long addr)
{
	unsigned char *new, *old;
	unsigned long ip = rec->ip;

	old = ftrace_nop_replace();
	new = ftrace_call_replace(ip, addr);

	return ftrace_modify_code(rec->ip, old, new);
}

int ftrace_update_ftrace_func(ftrace_func_t func)
{
	unsigned long ip = (unsigned long) &ftrace_call;
	unsigned char old[MCOUNT_INSN_SIZE], *new;
	int ret;

	memcpy(old, &ftrace_call, MCOUNT_INSN_SIZE);
	new = ftrace_call_replace(ip, (unsigned long)func);
	ret = ftrace_modify_code(ip, old, new);

	return ret;
}

int __init ftrace_dyn_arch_init(void *data)
{
	*(unsigned long *)data = 0;

	return 0;
}
