/*
 *  File Name	    : linux/drivers/i2c/busses/i2c-mp200.c
 *  Function	    : i2c
 *  Release Version : Ver 1.25
 *  Release Date    : 2009/07/22
 *
 *  Copyright (C) NEC Electronics Corporation 2007, 2008, 2009
 *
 *
 *  This program is free software;you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by Free
 * Softwere Foundation; either version 2 of License, or (at your option) any
 * later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 * with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/version.h>
#include <linux/i2c.h>
#include <linux/io.h>
#include <asm/irq.h>
#include <linux/uaccess.h>
#include <linux/errno.h>
#include <linux/sched.h>

#include <linux/device.h>
#include <linux/moduleparam.h>
#include <linux/interrupt.h>
#include <linux/ioctl.h>
#include <linux/platform_device.h>

#include <mach/pmu.h>
#include <mach/smu.h>

#include "i2c-mp200.h"

/* #define I2C_DEBUG 1 */

#ifdef I2C_DEBUG
#define DBG(level, format, arg...) printk(level format, ## arg)
#else
#define DBG(level, format, arg...) do { } while (0)
#endif

static int i2c_pending[I2C_NR];
static spinlock_t i2c_lock;
static spinlock_t i2c_irq_lock;

static int irq = 1;	/* polling mode flag(default=1 is interrupt mode) */

static int mp200_i2c_xfer(struct i2c_adapter *, struct i2c_msg[], int);
static int mp200_i2c_xbytes(struct i2c_adapter *, struct i2c_msg *, int);
static int mp200_i2c_suspend(struct platform_device *dev, pm_message_t state);
static int mp200_i2c_resume(struct platform_device *dev);


typedef struct {
	struct i2c_adapter	adap;
	wait_queue_head_t	i2c_wait;
	char			*membase;
	int			irq;
	unsigned int		ch;
	int			rstdev;
	int			clkdev;
	int			sclkdev;
	int			mst;
} mp200_i2c_device;


#ifndef CONFIG_I2C_MP200_DFC
#define CONFIG_I2C_MP200_DFC I2C_DFC_OFF
#endif


static i2c_ctrl_t mp200_i2c_ctrl = {
	.smc = CONFIG_I2C_MP200_SMC,	/* mode */
	.dfc = CONFIG_I2C_MP200_DFC,	/* digital filter */
};

static u32 mp200_i2c_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}


static struct i2c_algorithm mp200_i2c_algo = {
	.master_xfer = mp200_i2c_xfer,
	.smbus_xfer = NULL,
	.functionality = mp200_i2c_func,
};

static mp200_i2c_device mp200_i2c_devs[I2C_NR] = {
	[0] = {
		.adap = {
			.owner = THIS_MODULE,
			.name = "mp200 I2C1 adapter",
			.id = I2C_HW_MP200,
			.algo = &mp200_i2c_algo,
			.algo_data = &mp200_i2c_ctrl,
			.client_register = NULL,
			.client_unregister = NULL,
			.timeout = 100,		/* 1sec */
			.retries = 3,		/* 3times */
		},
		.membase = (void *)I2C_ADDR,
		.irq	= INT_IIC,
		.ch	= MP200_I2C_CH1,
		.rstdev = MP200_RESETDEVICE_IIC,
		.clkdev = MP200_CLOCKGATE_IIC_CLK,
		.sclkdev = MP200_CLOCKGATE_IIC_SCLK,
		.mst    = MP200_I2C_INTC_MST,
	},
#ifdef CONFIG_I2C_MP200_ENABLE_CH2
	[1] = {
		.adap = {
			.owner = THIS_MODULE,
			.name = "mp200 I2C2 adapter",
			.id = I2C_HW_MP200,
			.algo = &mp200_i2c_algo,
			.algo_data = &mp200_i2c_ctrl,
			.client_register = NULL,
			.client_unregister = NULL,
			.timeout = 100,		/* 1sec */
			.retries = 3,		/* 3times */
		},
		.membase = (void *)I2C2_ADDR,
		.irq	= INT_IIC2,
		.ch	= MP200_I2C_CH2,
		.rstdev = MP200_RESETDEVICE_IIC2,
		.clkdev = MP200_CLOCKGATE_IIC2_CLK,
		.sclkdev = MP200_CLOCKGATE_IIC2_SCLK,
		.mst    = MP200_I2C2_INTC_MST,
	},
#endif
};

static void mp200_i2c_enable_clock(mp200_i2c_device *i2c_dev)
{
	mp200_pmu_open_clockgate(i2c_dev->sclkdev);
	mp200_pmu_open_clockgate(i2c_dev->clkdev);
}

static void mp200_i2c_disable_clock(mp200_i2c_device *i2c_dev)
{
	mp200_pmu_close_clockgate(i2c_dev->clkdev);
	mp200_pmu_close_clockgate(i2c_dev->sclkdev);
}

static int mp200_i2c_get_clock_status(struct platform_device *dev)
{
	mp200_i2c_device *i2c_dev = platform_get_drvdata(dev);

	if (mp200_pmu_get_clockgate(i2c_dev->sclkdev) ||
		mp200_pmu_get_clockgate(i2c_dev->clkdev)) {
		return 1;
	}

	return 0;
}

static void mp200_i2c_setup_mode(struct i2c_adapter *adap)
{
	u16 bit = 0x0;
	i2c_ctrl_t *ctrl = (i2c_ctrl_t *) adap->algo_data;
	mp200_i2c_device *i2c_dev;

	i2c_dev = (mp200_i2c_device *)(i2c_get_adapdata(adap));

	DBG(KERN_INFO, "%s(): IICCL0: %04x (reg:0x%08x)\n",
		__func__,
		inw(i2c_dev->membase + I2C_OFS_IICCL0),
		(unsigned int)i2c_dev->membase+I2C_OFS_IICCL0);

	/* mode */
	if (ctrl->smc == I2C_SMC_HIGH_SPEED) {
		/* High-speed mode */
		bit |= I2C_BIT_SMC0;
	}

	/* digital filter */
	if (ctrl->dfc == I2C_DFC_ON) {
		/* digital filter ON */
		bit |= I2C_BIT_DFC0;
	}

	outw(bit, i2c_dev->membase + I2C_OFS_IICCL0);

	DBG(KERN_INFO, "%s(): IICCL0: %04x (reg:0x%08x)\n",
		__func__,
		inw(i2c_dev->membase + I2C_OFS_IICCL0),
		(unsigned int)i2c_dev->membase + I2C_OFS_IICCL0);
}

static int mp200_i2c_wait_sp(mp200_i2c_device *i2c_dev)
{
	int timeout = 200;

	/* Stop condition */
	spin_lock(&i2c_lock);
	outw((inw(i2c_dev->membase + I2C_OFS_IICC0) | I2C_BIT_SPT0),
		i2c_dev->membase + I2C_OFS_IICC0);
	spin_unlock(&i2c_lock);

	while ((inw(i2c_dev->membase + I2C_OFS_IICSE0) & I2C_BIT_SPD0) == 0) {
		if (0 == timeout--) {
			DBG(KERN_INFO,
				"%s(): could not detect stop condition\n",
				__func__);
			return -EBUSY;
		}
		udelay(1);
	}
	DBG(KERN_INFO, "%s(): stop:: con: %04x, stat: %04x\n",
		__func__,
		inw(i2c_dev->membase + I2C_OFS_IICC0),
		inw(i2c_dev->membase + I2C_OFS_IICSE0));

	return 0;
}

static void mp200_i2c_reset(struct i2c_adapter *adap)
{
	mp200_i2c_device *i2c_dev;

	i2c_dev = (mp200_i2c_device *)(i2c_get_adapdata(adap));
	/* I2C disable */
	if (inw(i2c_dev->membase + I2C_OFS_IICC0) & I2C_BIT_IICE0) {
#if 1
		if ((inw(i2c_dev->membase + I2C_OFS_IICCL0)
				& (I2C_BIT_CLD0 | I2C_BIT_DAD0))
				!= (I2C_BIT_CLD0 | I2C_BIT_DAD0)) {

			mp200_i2c_wait_sp(i2c_dev);
			/* what about case error returned. */
		}
#endif
		spin_lock(&i2c_lock);
		outw(0, i2c_dev->membase + I2C_OFS_IICC0);
		spin_unlock(&i2c_lock);
	}
	/* transfer mode set */
	mp200_i2c_setup_mode(adap);

	/* Refuse transaction */
	/* Can write STCEN at not trns other i2c.*/
	outw((I2C_BIT_STCEN | I2C_BIT_IICRSV),
		i2c_dev->membase + I2C_OFS_IICF0);

	/* I2C enable, 9bit interrupt mode */
	spin_lock(&i2c_lock);
	outw((I2C_BIT_IICE0 | I2C_BIT_WTIM0), i2c_dev->membase + I2C_OFS_IICC0);
	spin_unlock(&i2c_lock);

	DBG(KERN_INFO, "%s()_: con: %04x, stat: %04x (reg:0x%08x,0x%08X)\n",
		__func__,
		inw(i2c_dev->membase + I2C_OFS_IICC0),
		inw(i2c_dev->membase + I2C_OFS_IICSE0),
		(unsigned int)i2c_dev->membase + I2C_OFS_IICC0,
		(unsigned int)i2c_dev->membase + I2C_OFS_IICSE0);
}

/*
 * Wait until I2C bus is free
 */
static int mp200_i2c_wait_for_bb(struct i2c_adapter *adap)
{
	mp200_i2c_device *i2c_dev;
	int status;
	int timeout = adap->timeout;

	i2c_dev = (mp200_i2c_device *)(i2c_get_adapdata(adap));

	/* wait until I2C bus free */
	while ((inw(i2c_dev->membase + I2C_OFS_IICF0) & I2C_BIT_IICBSY)
			&& timeout--) {
		schedule_timeout_uninterruptible(1);
	}

	status = (timeout <= 0);		/* set 0 or 1 */
	if (status)
		DBG(KERN_INFO, "%s(): Timeout, I2C bus is busy\n", __func__);

	return status;
}

/*
 * Wait until INT_I2C handler will be interrupted
 */
static int mp200_i2c_wait_for_pin(struct i2c_adapter *adap, u16 *status)
{
	mp200_i2c_device *i2c_dev;
	int retries = adap->retries;
	unsigned long timeout;
	int pending = 0;
	sigset_t oldset;
	sigset_t newset;

	DECLARE_WAITQUEUE(wait, current);

	i2c_dev = (mp200_i2c_device *)(i2c_get_adapdata(adap));

	if (irq) {
		add_wait_queue(&i2c_dev->i2c_wait, &wait);
		spin_lock_irq(&current->sighand->siglock);
		oldset = current->blocked;
		newset = current->blocked;
		spin_unlock_irq(&current->sighand->siglock);
		sigaddset(&newset, SIGINT);
		sigaddset(&newset, SIGKILL);
		sigprocmask(SIG_BLOCK, &newset, NULL);
	}

	do {
		/* interrupt mode */
		if (irq) {
			timeout = adap->timeout;
			current->state = TASK_UNINTERRUPTIBLE;

			spin_lock_irq(&i2c_irq_lock);
			pending = i2c_pending[i2c_dev->ch];
			i2c_pending[i2c_dev->ch] = 0;
			spin_unlock_irq(&i2c_irq_lock);

			*status = inw(i2c_dev->membase + I2C_OFS_IICSE0);
			DBG(KERN_INFO,
			 "%s(): con: %04x, stat: %04x (reg:0x%08x,0x%08x)\n",
			 __func__,
			 inw(i2c_dev->membase + I2C_OFS_IICC0),
			 *status,
			 (unsigned int)i2c_dev->membase + I2C_OFS_IICC0,
			 (unsigned int)i2c_dev->membase + I2C_OFS_IICC0);

			if (pending)
				break;

			schedule_timeout(timeout);
		}
		/* polling mode */
		else {
			timeout = adap->timeout * 10000;
			do {
				udelay(10);

				/* Check INT_I2C/INT_I2C2 bit */
				if (i2c_dev->mst
					& inl(MP200_I2C_INTC_IT0_RAW)) {

					outl(i2c_dev->mst,
						MP200_I2C_INTC_IT0_IIR);
					pending = 1;
					break;
				}
			} while (--timeout);

			*status = inw(i2c_dev->membase + I2C_OFS_IICSE0);
			DBG(KERN_INFO, "%s(): con: %04x, stat: %04x\n",
				__func__,
				inw(i2c_dev->membase + I2C_OFS_IICC0), *status);

			if (pending)
				break;

		}

	} while ((*status & I2C_BIT_MSTS0) && retries--);

	/* interrupt mode */
	if (irq) {
		sigprocmask(SIG_SETMASK, &oldset, NULL);
		current->state = TASK_RUNNING;
		remove_wait_queue(&i2c_dev->i2c_wait, &wait);
	}

	/* retry counter is expired */
	if (retries <= 0)
		DBG(KERN_INFO, "%s(): Timeout\n", __func__);


	return (retries <= 0) ? -1 : 0;
}

/*
 * master_xfer Interface
 */
static int mp200_i2c_xfer(struct i2c_adapter *adap, struct i2c_msg *msgs,
	int num)
{
	int i;
	int status = 0;
	mp200_i2c_device *i2c_dev;

	DBG(KERN_INFO, "%s(): msgs: %d\n", __func__, num);

	i2c_dev = (mp200_i2c_device *)(i2c_get_adapdata(adap));

	/* Check invalid param in i2c_msg */
	if (msgs == NULL)
		return -EINVAL;

	for (i = 0; i < num; i++) {
		if (msgs[i].buf == NULL)
			return -EINVAL;

	}

	/* I2C macro Enable */
	mp200_i2c_enable_clock(i2c_dev);
	/* I2C transfer mode set */
	mp200_i2c_reset(adap);

	status = mp200_i2c_wait_for_bb(adap);
	if (status) {
		DBG(KERN_INFO, "%s(): %s: bus busy\n", __func__, adap->name);

		if (!(inw(i2c_dev->membase + I2C_OFS_IICSE0)
				& I2C_BIT_MSTS0)) {
			/* Slave mode -> Error */
			mp200_i2c_disable_clock(i2c_dev);
			return -EBUSY;
		}
		DBG(KERN_INFO,
			"%s(): %s is already active. Stop Condition...\n",
			__func__, adap->name);
		/* Stop condition */
		spin_lock(&i2c_lock);
		outw((inw(i2c_dev->membase + I2C_OFS_IICC0) | I2C_BIT_SPT0),
			i2c_dev->membase + I2C_OFS_IICC0);
		spin_unlock(&i2c_lock);

		status = mp200_i2c_wait_for_bb(adap);
		if (status) {
			DBG(KERN_INFO,
				"%s(): %s is already active. Reset...\n",
				__func__, adap->name);
			/* I2C Reset */
			mp200_i2c_reset(adap);

			if (mp200_i2c_wait_for_bb(adap)) {
				DBG(KERN_INFO,
				"%s(): %s is already active. Reset failed...\n",
				__func__, adap->name);

				mp200_i2c_disable_clock(i2c_dev);
				return -EREMOTEIO;
			}
		}
	}

	for (i = 0; i < num; i++) {
		DBG(KERN_INFO,
			"%s(): msg: %d, addr: 0x%04x, len: %d, flags: 0x%x\n",
			__func__, i, msgs[i].addr, msgs[i].len, msgs[i].flags);

		/* Transfer msgs[] */
		status = mp200_i2c_xbytes(adap, &msgs[i], (i == (num - 1)));

		DBG(KERN_INFO, "%s(): status : %d\n", __func__, status);
		if (status < 0) {
			/* transfer error */
			break;
		}
	}

	/* I2C transfer completed */
	if (status >= 0) {
		/* the number of i2c_msg transfered */
		status = i;
	}

	DBG(KERN_INFO, "%s(): status : %d\n", __func__, status);

	/* I2C macro Disable */
	mp200_i2c_disable_clock(i2c_dev);
	/* if after stop clock then register all zero. */

	return status;
}

/*
 * master_xfer transaction(Low level)
 */
static int mp200_i2c_xbytes(struct i2c_adapter *adap, struct i2c_msg *msg,
	int stop)
{
	u16 status;
	u8 data = 0;
	int count = 0;
	int timeout;
	mp200_i2c_device *i2c_dev = (mp200_i2c_device *)i2c_get_adapdata(adap);

	DBG(KERN_INFO, "%s(): addr: 0x%04x, len: %d, flags: 0x%x, stop: %d\n",
		__func__, msg->addr, msg->len, msg->flags, stop);

	/* interrupt mode */
	if (irq) {
		spin_lock_irq(&i2c_irq_lock);
		i2c_pending[i2c_dev->ch] = 0;
		spin_unlock_irq(&i2c_irq_lock);
	}

	/* Start condition */
	spin_lock(&i2c_lock);
	outw((inw(i2c_dev->membase + I2C_OFS_IICC0) & ~I2C_BIT_ACKE0)
		| I2C_BIT_WTIM0, i2c_dev->membase + I2C_OFS_IICC0);
	outw((inw(i2c_dev->membase + I2C_OFS_IICC0) | I2C_BIT_STT0),
		i2c_dev->membase + I2C_OFS_IICC0);
	spin_unlock(&i2c_lock);

	/* wait for completion of Start condition */
	timeout = 10;
	do {
		udelay(MP200_I2C_WAIT); /* 1us wait */
		DBG(KERN_INFO, "%s(): con: %04x, flg: %04x\n", __func__,
			inw(i2c_dev->membase + I2C_OFS_IICC0),
			inw(i2c_dev->membase + I2C_OFS_IICF0));
		if (((inw(i2c_dev->membase + I2C_OFS_IICF0)
			& I2C_BIT_STCF) == 0)
			&& (inw(i2c_dev->membase + I2C_OFS_IICSE0)
			& I2C_BIT_MSTS0)) {
			break;
		}
		if (--timeout == 0)
			return -EBUSY;

	} while (1);
	DBG(KERN_INFO, "%s(): addr: %04x, flags: %04x\n", __func__, msg->addr,
		msg->flags);
	DBG(KERN_INFO, "%s(): addrdir: %04x\n", __func__,
		(msg->addr << I2C_DIR_SHIFT)
		| ((msg->flags & I2C_M_RD) ? 1 : 0));

	/* Send slave address and R/W type */
	outw((msg->addr << I2C_DIR_SHIFT)
		| ((msg->flags & I2C_M_RD) ? 1 : 0),
		i2c_dev->membase + I2C_OFS_IIC0);

	/* Wait for transaction */
	if (mp200_i2c_wait_for_pin(adap, &status)) {
		mp200_i2c_reset(adap);
		return -EREMOTEIO;
	}

	/* Arbitration */
	if (status & I2C_BIT_ALD0) {
		mp200_i2c_reset(adap);
		return -EREMOTEIO;
	}
	/* Extension mode or Slave mode */
	else if (status & (I2C_BIT_EXC0 | I2C_BIT_COI0)
			|| !(status & I2C_BIT_MSTS0)) {
		mp200_i2c_reset(adap);
		return -EREMOTEIO;
	}
	/* Read transaction */
	else if (!(status & I2C_BIT_TRC0)) {
		/* msg->flags is Write type */
		if (!(msg->flags & I2C_M_RD)) {
			DBG(KERN_INFO, "%s(): %s r/w hardware fault.\n",
				__func__, adap->name);
			mp200_i2c_reset(adap);
			return -EREMOTEIO;
		}

		/* Recieved No ACK (result of setting slave address and R/W) */
		if (!(status & I2C_BIT_ACKD0)) {
			DBG(KERN_INFO, "%s(): %s recieved No ACK.\n",
				__func__, adap->name);
			/* Stop condition */
			mp200_i2c_wait_sp(i2c_dev);
			return -EREMOTEIO;
		}

		/* 8bit interrupt mode */
		spin_lock(&i2c_lock);
		outw((inw(i2c_dev->membase + I2C_OFS_IICC0) & ~I2C_BIT_WTIM0)
			| I2C_BIT_ACKE0, i2c_dev->membase + I2C_OFS_IICC0);
		outw((inw(i2c_dev->membase + I2C_OFS_IICC0) & ~I2C_BIT_WTIM0)
			| I2C_BIT_WREL0, i2c_dev->membase + I2C_OFS_IICC0);
		spin_unlock(&i2c_lock);

		/* Wait for transaction */
		if (mp200_i2c_wait_for_pin(adap, &status)) {
			mp200_i2c_reset(adap);
			return -EREMOTEIO;
		}
	}

	do {
		/* Arbitration */
		if (status & I2C_BIT_ALD0) {
			mp200_i2c_reset(adap);
			return -EREMOTEIO;
		}
		/* Extension mode or Slave mode */
		else if (status & (I2C_BIT_EXC0 | I2C_BIT_COI0)
				|| !(status & I2C_BIT_MSTS0)) {
			mp200_i2c_reset(adap);
			return -EREMOTEIO;
		}
		/* Read transaction */
		else if (!(status & I2C_BIT_TRC0)) {
			/* msg->flags is Write type */
			if (!(msg->flags & I2C_M_RD)) {
				DBG(KERN_INFO, "%s(): %s r/w hardware fault.\n",
					__func__, adap->name);
				mp200_i2c_reset(adap);
				return -EREMOTEIO;
			}

			/* Read transaction is completed without any errors */
			if (count == msg->len)
				break;

			/* read data */
			data = inw(i2c_dev->membase + I2C_OFS_IIC0);
			DBG(KERN_INFO, "%s(): read: data: %04x\n", __func__,
				data);
			if (count < msg->len) {
				msg->buf[count++] = data;
			} else if (msg->len != 0) {
				/* recieve error */
				DBG(KERN_INFO, "%s(): %s rcv count mismatch.\n",
					__func__, adap->name);
				break;
			}

			if (count < msg->len) {
				spin_lock(&i2c_lock);
				outw((inw(i2c_dev->membase + I2C_OFS_IICC0)
					| I2C_BIT_WREL0),
					i2c_dev->membase + I2C_OFS_IICC0);
				spin_unlock(&i2c_lock);
			} else {
				spin_lock(&i2c_lock);
				outw((inw(i2c_dev->membase + I2C_OFS_IICC0)
					& ~I2C_BIT_ACKE0) | I2C_BIT_WTIM0,
					i2c_dev->membase + I2C_OFS_IICC0);
				outw(inw(i2c_dev->membase + I2C_OFS_IICC0)
					| I2C_BIT_WREL0,
					i2c_dev->membase + I2C_OFS_IICC0);
				spin_unlock(&i2c_lock);
			}

		/* Write transaction */
		} else if (status & I2C_BIT_TRC0) {
			/* msg->flags is Read type */
			if ((msg->flags & I2C_M_RD)) {
				DBG(KERN_INFO, "%s(): %s r/w hardware fault.\n",
					__func__, adap->name);
				mp200_i2c_reset(adap);
				return -EREMOTEIO;
			}

			/* Recieved No ACK (result of setting slave address
			 * and R/W)
			 */
			if (!(status & I2C_BIT_ACKD0)) {
				DBG(KERN_INFO, "%s(): %s recieved No ACK.\n",
					__func__, adap->name);
				/* Stop condition */
				mp200_i2c_wait_sp(i2c_dev);
				return -EREMOTEIO;
			}

			if (count < msg->len) {
				data = msg->buf[count++];
			} else if (count == msg->len) {
				/* send transaction is completed without any
				 * errors
				 */
				break;
			} else if (msg->len != 0) {
				/* send error */
				DBG(KERN_INFO, "%s(): %s xmt count mismatch.\n",
					__func__, adap->name);
				break;
			}

			/* write data */
			DBG(KERN_INFO, "%s(): write: data: %04x\n", __func__,
				data);
			outw(data, i2c_dev->membase + I2C_OFS_IIC0);

		/* not reached ? */
		} else {
			DBG(KERN_INFO, "%s(): con: %04x, stat: %04x\n",
				__func__, inw(i2c_dev->membase + I2C_OFS_IICC0),
				inw(i2c_dev->membase + I2C_OFS_IICSE0));
		}

		/* Wait for R/W transaction */
		if (mp200_i2c_wait_for_pin(adap, &status)) {
			mp200_i2c_reset(adap);
			return -EREMOTEIO;
		}
	} while (count <= msg->len);

	DBG(KERN_INFO, "%s(): count: %d, len: %d, stop %d\n", __func__, count,
		msg->len, stop);

	/* Stop condition */
	if ((count >= msg->len) && stop) {
		spin_lock(&i2c_lock);
		outw((inw(i2c_dev->membase + I2C_OFS_IICC0) | I2C_BIT_SPT0),
			i2c_dev->membase + I2C_OFS_IICC0);
		spin_unlock(&i2c_lock);
		timeout = 200;
		while ((inw(i2c_dev->membase + I2C_OFS_IICSE0)
				& I2C_BIT_SPD0) == 0) {
			if (0 == timeout--) {
				DBG(KERN_INFO,
				"%s(): could not detect stop condition\n",
				__func__);
				return -EBUSY;
			}
			udelay(1);
		}
		DBG(KERN_INFO, "%s(): stop:: con: %04x, stat: %04x\n",
			__func__, inw(i2c_dev->membase + I2C_OFS_IICC0),
			inw(i2c_dev->membase + I2C_OFS_IICSE0));
	}

	return (count > 0) ? (msg->len ? count : 0) : count;
}

static int mp200_i2c_add_bus(struct i2c_adapter *adap)
{
	mp200_i2c_device *i2c_dev =
		(mp200_i2c_device *)(i2c_get_adapdata(adap));

	DBG(KERN_INFO, "%s(): hw routines for %s registered.\n", __func__,
		adap->name);

	i2c_add_numbered_adapter(adap);

	mp200_i2c_enable_clock(i2c_dev);
	mp200_pmu_unreset_device(i2c_dev->rstdev);
	mp200_i2c_reset(adap);
	mp200_i2c_disable_clock(i2c_dev);

	return 0;
}

static int mp200_i2c_del_bus(struct i2c_adapter *adap)
{
	int res;

	res = i2c_del_adapter(adap);
	if (res < 0)
		return res;

	DBG(KERN_INFO, "%s(): adapter unregistered: %s\n", __func__,
		adap->name);

	return 0;
}


static irqreturn_t mp200_i2c_handler(int this_irq, void *dev_id,
	struct pt_regs *regs)
{
	mp200_i2c_device *i2c_dev = dev_id;

	spin_lock(&i2c_irq_lock);
	i2c_pending[i2c_dev->ch] = 1;
	spin_unlock(&i2c_irq_lock);

	DBG(KERN_INFO, "%s(): in interrupt handler\n", __func__);

	wake_up(&i2c_dev->i2c_wait);

	return IRQ_HANDLED;
}


static int mp200_i2c_hw_resrc_init(struct platform_device *dev)
{
	mp200_i2c_device *i2c_dev = platform_get_drvdata(dev);
	DBG(KERN_INFO, "%s(): resrc Initializing.\n", __func__);

	if (i2c_dev->ch == MP200_I2C_CH1) {
		outl((inl(SMU_DIVIICSCLK) & 0x00ff0000) | I2C_DIVIIC_VAL1,
			SMU_DIVIICSCLK);
		if (irq) {
			if (request_irq(i2c_dev->irq, (void *)mp200_i2c_handler,
					0, "mp200_i2c",
					&mp200_i2c_devs[0]) < 0) {
				return -ENODEV;
			}
		}
	}
#ifdef CONFIG_I2C_MP200_ENABLE_CH2
	else
	if (i2c_dev->ch == MP200_I2C_CH2) {
		outl((inl(SMU_DIVIICSCLK) & 0x000000ff)
			| I2C_DIVIIC_VAL2, SMU_DIVIICSCLK);
		if (irq) {
			if (request_irq(i2c_dev->irq, (void *)mp200_i2c_handler,
					0, "mp200_i2c",
					&mp200_i2c_devs[1]) < 0) {
				return -ENODEV;
			}
		}
	}
#endif

	return 0;
}

static void mp200_i2c_release(struct platform_device *dev)
{
	mp200_i2c_device *i2c_dev = platform_get_drvdata(dev);

	if (irq)
		free_irq(i2c_dev->irq, &mp200_i2c_devs[i2c_dev->ch]);

}


static int mp200_i2c_suspend(struct platform_device *dev, pm_message_t state)
{
	switch (state.event) {
	case DEV_SUSPEND_IDLE_1:
	case PM_EVENT_SUSPEND:
		if (mp200_i2c_get_clock_status(dev))
			return -EBUSY;
		break;
	default:
		break;
	}

	return 0;
}

static int mp200_i2c_resume(struct platform_device *dev)
{
	return 0;
}


/*
 * this function is called when this module is loaded.
 */
static int mp200_i2c_probe(struct platform_device *dev)
{
	DBG(KERN_INFO, "Starting mp200_i2c.\n");

	if (dev->id > I2C_NR)
		return -ENODEV;

	platform_set_drvdata(dev, &mp200_i2c_devs[dev->id]);
	i2c_set_adapdata(&mp200_i2c_devs[dev->id].adap,
				&mp200_i2c_devs[dev->id]);

	if (mp200_i2c_hw_resrc_init(dev) < 0)
		return -ENODEV;

	init_waitqueue_head(&mp200_i2c_devs[dev->id].i2c_wait);

	mp200_i2c_devs[dev->id].adap.nr = dev->id;
	if (mp200_i2c_add_bus(&mp200_i2c_devs[dev->id].adap) < 0)
		return -ENODEV;

	return 0;
}

static int mp200_i2c_remove(struct platform_device *dev)
{
	mp200_i2c_device *i2c_dev = platform_get_drvdata(dev);

	DBG(KERN_INFO, "%s(): ch=%d.\n", __func__, i2c_dev->ch);

	mp200_i2c_del_bus(&mp200_i2c_devs[i2c_dev->ch].adap);
	mp200_i2c_release(dev);

	return 0;
}

static struct platform_driver mp200_i2c_driver = {
	.probe = mp200_i2c_probe,
	.remove = mp200_i2c_remove,
	.suspend = mp200_i2c_suspend,
	.resume = mp200_i2c_resume,
	.driver = {
		.name = "i2c",
		.owner = THIS_MODULE
	}
};


static int __init mp200_i2c_init(void)
{
	spin_lock_init(&i2c_lock);
	spin_lock_init(&i2c_irq_lock);

	return platform_driver_register(&mp200_i2c_driver);
}

static void __exit mp200_i2c_exit(void)
{
	platform_driver_unregister(&mp200_i2c_driver);
}

module_param(irq, int, 0);

MODULE_LICENSE("GPL");

module_init(mp200_i2c_init);
module_exit(mp200_i2c_exit);
