/*
 * atmel49xx_flash.c
 *
 * The map file for Boot Flash.
 *
 * Author: Force Computers (India) PVT. LTD.
 *         Soumya K.S <soumya.ks@fci.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 *  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>


#define FLASH_PHYS_ADDR 0xFFF00000	/* Boot Flash Area */

#define FLASH_SIZE 0x100000

/*
 * Two partitions have been defined so far:
 *
 * 1. Partition 1 (0xFFF00000 - 0xFFFFFFFF)  -- 1.00MB
 */

#define FLASH_PARTITION0_ADDR 0x00000000
#define FLASH_PARTITION0_SIZE 0x00100000

map_word atmel49xx_read8(struct map_info *map, unsigned long ofs)
{
	map_word val;
	val.x[0] = readb((void __iomem *) (map->map_priv_1 + ofs));
	return val;
}

map_word atmel49xx_read16(struct map_info *map, unsigned long ofs)
{
	map_word val;
	val.x[0] = readw((void __iomem *) (map->map_priv_1 + ofs));
	return val;
}

map_word atmel49xx_read32(struct map_info *map, unsigned long ofs)
{
	map_word val;
	val.x[0] = readl((void __iomem *) (map->map_priv_1 + ofs));
	return val;
}

void atmel49xx_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void __iomem *) (map->map_priv_1 + from), len);
}

void atmel49xx_write8(struct map_info *map, map_word val, unsigned long adr)
{
	writeb(val.x[0], (void __iomem *) (map->map_priv_1 + adr));
	mb();
}

void atmel49xx_write16(struct map_info *map, map_word val, unsigned long adr)
{
	writew(val.x[0], (void __iomem *) (map->map_priv_1 + adr));
	mb();
}

void atmel49xx_write32(struct map_info *map, map_word val, unsigned long adr)
{
	writel(val.x[0], (void __iomem *) (map->map_priv_1 + adr));
	mb();
}

void atmel49xx_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void __iomem *) (map->map_priv_1 + to), from, len);
}

struct map_info atmel49xx_map = {
	.name = "Boot flash device",
	.phys = FLASH_PHYS_ADDR,
	.size = FLASH_SIZE,
	.bankwidth = 1,
	.read = atmel49xx_read8,
	.copy_from = atmel49xx_copy_from,
	.write = atmel49xx_write8,
	.copy_to = atmel49xx_copy_to
};

struct mtd_partition atmel49xx_parts[] = {
	{
		name	: "BIOS Image",
		offset	: FLASH_PARTITION0_ADDR,
		size	: FLASH_PARTITION0_SIZE
	},
};


#define PARTITION_COUNT (sizeof(atmel49xx_parts)/sizeof(struct mtd_partition))

static struct mtd_info *mymtd;

#define	T_BLOCK_MINUS15		0xFFB00002

void fwh_init (void)
{
	unsigned char *fwh_tblocks;
	unsigned int count=0;

	/*
	** set sectors to be in full access lock down.  This means that the
	** value of the lock bit cannot be changed, i.e. the chip is always
	** configured to allow read and write access.
	*/

	for (count=0; count<=0xF0000; count+=0x10000) {
		fwh_tblocks = (unsigned char*)ioremap_nocache ((T_BLOCK_MINUS15+count),4);
		*fwh_tblocks = 0x02;
		iounmap ((unsigned char*)fwh_tblocks);
	}
}

int __init init_atmel49xx(void)
{
	/*unlock blocks*/
	fwh_init ();

	printk(KERN_NOTICE "Atmel Boot flash device: %x at %x\n",
			FLASH_SIZE, FLASH_PHYS_ADDR);

	atmel49xx_map.map_priv_1 = (unsigned long)ioremap( FLASH_PHYS_ADDR, FLASH_SIZE ) ;

	if (!atmel49xx_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	mymtd = do_map_probe("jedec_probe", &atmel49xx_map);
	if (mymtd) {
		mymtd->owner = THIS_MODULE;

		/*
		** since the chip cannot be locked make sure the functions are
		** not configured so that -EOPNOTSUPP is returned when user
		** attempts to lock an area
		*/
		mymtd->lock = NULL;
		mymtd->unlock = NULL;
		add_mtd_partitions(mymtd, atmel49xx_parts, PARTITION_COUNT);
		printk(KERN_NOTICE "Atmel Boot flash device initialized\n");
		return 0;
	}

	iounmap((void *)atmel49xx_map.map_priv_1);

	return -ENXIO;
}

static void __exit cleanup_atmel49xx(void)
{
	if (mymtd) {
		del_mtd_partitions(mymtd);
		map_destroy(mymtd);
	}
	if (atmel49xx_map.map_priv_1) {
		iounmap((void *)atmel49xx_map.map_priv_1);
		atmel49xx_map.map_priv_1 = 0;
	}
}

module_init(init_atmel49xx);
module_exit(cleanup_atmel49xx);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("K S Soumya. <Soumya.KS@fci.com>");
MODULE_DESCRIPTION("MTD map driver for Atmel Flash chip");
