/* drivers/power/mp200_battery.c
 *
 * Power supply driver for the mp200 series
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/module.h>
#include <linux/err.h>
#include <linux/platform_device.h>
#include <linux/power_supply.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/timer.h>
#include <linux/io.h>
#include <mach/pwc.h>
#include <mach/smu.h>
#include <mach/pm.h>

#include "mp200_battery.h"

//#define MP200_BATTERY_DEBUG

#define MP200_BATTERY_NAME "mp200-battery"

#ifdef MP200_BATTERY_DEBUG
#define DEBUG_PRINT(FMT, ARGS...) printk(KERN_INFO "%s: " FMT, __func__, ##ARGS)
#else
#define DEBUG_PRINT(FMT, ARGS...)
#endif

/* battery watch timer sec */
#ifdef MP200_BATTERY_DEBUG
#define MP200_CHARGE_CHGFALL_WATCH_TIME (5 * 1000)	/* 5sec */
#else
#define MP200_CHARGE_CHGFALL_WATCH_TIME (30 * 1000)	/* 30sec */
#endif


#ifdef CONFIG_MP200_BATTERY_VOLTAGE_LIMIT
#if CONFIG_MP200_BATTERY_VOLTAGE_LIMIT <= 0x1F
#define CHARGING_VOLTAGE_LIMIT CONFIG_MP200_BATTERY_VOLTAGE_LIMIT
#else
#define CHARGING_VOLTAGE_LIMIT VCHG_BAT_LIION
#endif
#else
#define CHARGING_VOLTAGE_LIMIT VCHG_BAT_LIION
#endif

#ifdef CONFIG_MP200_BATTERY_CURRENT_LIMIT
#if CONFIG_MP200_BATTERY_CURRENT_LIMIT <= 0x3F
#define CHARGING_CURRENT_LIMIT CONFIG_MP200_BATTERY_CURRENT_LIMIT
#else
#define CHARGING_CURRENT_LIMIT ICHG_BAT_20MA
#endif
#else
#define CHARGING_CURRENT_LIMIT ICHG_BAT_20MA
#endif

struct timer_list mp200_battery_watch_timer;
static struct workqueue_struct *mp200_battery_workqueue;
static struct work_struct mp200_battery_work;

/* Battery charger limit */
static int voltage_limit;
static int current_limit;


enum charger_type_t {
	CHARGER_NONE = 0,
	CHARGER_BATTERY,
	CHARGER_USB_CG,
	CHARGER_USB_PC100,
	CHARGER_USB_PC500,
	CHARGER_AC
};


struct mp200_battery_data {
	spinlock_t lock;

	struct power_supply battery;
	struct power_supply usb;
	struct power_supply ac;

	enum charger_type_t charger;
	enum charger_type_t battery_det;
	enum charger_type_t usb_det;
	enum charger_type_t ac_det;

	unsigned int level;
	int low_mode;

// BEGIN eflow-added
	unsigned int voltage_now;
	unsigned int charge_now;
// END eflow-added
};

/* temporary variable used between mp200_battery_probe() *
 * and mp200_battery_open()                              */
static struct mp200_battery_data *battery_data;

static enum power_supply_property mp200_battery_props[] = {
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_HEALTH,
	POWER_SUPPLY_PROP_PRESENT,
	POWER_SUPPLY_PROP_TECHNOLOGY,
// BEGIN eflow-added
	POWER_SUPPLY_PROP_VOLTAGE_NOW,
	POWER_SUPPLY_PROP_CHARGE_NOW,
// END eflow-added
	POWER_SUPPLY_PROP_CAPACITY,
};

static enum power_supply_property mp200_power_props[] = {
	POWER_SUPPLY_PROP_ONLINE,
};

/* read ADC CH3 */
static int adc_ch3_read(void)
{
	int adcresh = 0, adcresl = 0;
	unsigned int data;

	pwc_write(DA9052_EVENTB_REG, 0x20, 0x20);
	pwc_reg_write(DA9052_ADCMAN_REG, 0x13);

	do {	/* or wait for INT_PWC_E_ADC_EOM interrupt */
		pwc_read(DA9052_ADCMAN_REG, &data);
	} while (data & 0x10);


	pwc_read(DA9052_ADCRESH_REG, &adcresh);
	pwc_read(DA9052_ADCRESL_REG, &adcresl);

	return ((adcresh & 0xff) << 2) + (adcresl & 0x3);
}

// BEGIN eflow-added
/* read VDDRES */
static unsigned int adc_vdd_read(void)
{
	unsigned int data;
	unsigned int vdd;	// [mV]

	pwc_read(DA9052_VDDRES_REG, &data);
	vdd = data * 2000 / 255 + 2500;		// 0x00:2.5V --- 0xff:4.5V

	return vdd;
}
// END eflow-added

/* battery stats. on/off line, battery level */
static void mp200_battery_watch(void)
{
	int battery_adc = 0;
	int current_vbat = 0;
	int limit_bat;

	battery_adc = adc_ch3_read();

	if (battery_adc == 0) {
		if (battery_data->battery_det == CHARGER_BATTERY) {
			battery_data->battery_det = CHARGER_NONE;
			battery_data->level = 0;
		}
	} else {
		if (battery_data->battery_det == CHARGER_NONE)
			battery_data->battery_det = CHARGER_BATTERY;

		limit_bat = 3650 + voltage_limit * 25;	/* mV */

		/* ADC=[(VBAT-2.5V) * 0.5] * 1023 */
		current_vbat = (battery_adc * 4 + 5115) * 500 / 1023;	/* mV */
		DEBUG_PRINT("battery vbat %d\n", current_vbat);
		if (current_vbat < 3600)
			battery_data->level = 0;
		else
			battery_data->level = (current_vbat - 3600) *
				 100 / (limit_bat - 3600);
	}

// BEGIN eflow-added
	battery_data->charge_now = current_vbat;
	battery_data->voltage_now = adc_vdd_read();
// END eflow-added
}

/* charging stats. on/off line, battery level */
static unsigned int mp200_charger_get_status(void)
{
	unsigned int status;

	pwc_read(DA9052_STATUSA_REG, &status);
	DEBUG_PRINT("STATUSA 0x%02x\n", status);

	mp200_battery_watch();

	if (status & 0x08)
		battery_data->ac_det = CHARGER_AC;
	else
		battery_data->ac_det = CHARGER_NONE;

	if (status & 0x10) /* not support USB PC 500mA & USB CG */
		battery_data->usb_det = CHARGER_USB_PC100;
	else
		battery_data->usb_det = CHARGER_NONE;

	if (battery_data->ac_det == CHARGER_AC)
		battery_data->charger = CHARGER_AC;
	else if (battery_data->usb_det == CHARGER_USB_CG)
		battery_data->charger = CHARGER_USB_CG;
	else if (battery_data->usb_det == CHARGER_USB_PC500)
		battery_data->charger = CHARGER_USB_PC500;
	else if (battery_data->battery_det == CHARGER_BATTERY)
		battery_data->charger = CHARGER_BATTERY;
	else if (battery_data->usb_det == CHARGER_USB_PC100)
		battery_data->charger = CHARGER_USB_PC100;

	return status;
}


static void low_speed_mode(void)
{
	if ((inl(SMU_CLK_MODE_SEL) & 0x00000F00) == (2 << 8))
		return;	/* now mode normal B */

	pm_change_normalB();
}

static void full_speed_mode(void)
{
	if ((inl(SMU_CLK_MODE_SEL) & 0x00000F00) == (1 << 8))
		return; /* now mode normal A */

	pm_change_normalA();
}


static void mp200_charger_update_status(struct work_struct *work)
{
	enum charger_type_t old_charger = battery_data->charger;
	enum charger_type_t old_usb = battery_data->usb_det;

	mp200_charger_get_status();

	if ((old_usb != CHARGER_USB_PC100) &&
		 (battery_data->usb_det == CHARGER_USB_PC100)) {
		DEBUG_PRINT("ISET_USB 100mA \n");
		pwc_write(DA9052_ISET_REG, 0x03, 0x0f);
	} else if ((old_usb == CHARGER_USB_PC100) &&
		(battery_data->usb_det == CHARGER_NONE)) {
		DEBUG_PRINT("ISET_USB 500mA \n");
		pwc_write(DA9052_ISET_REG, 0x08, 0x0f);
	}

	if (old_charger != battery_data->charger) {

		power_supply_changed(&battery_data->battery);
		power_supply_changed(&battery_data->usb);
		power_supply_changed(&battery_data->ac);

		if (battery_data->charger == CHARGER_AC) {
			DEBUG_PRINT("ISET_BUCK 900mA\n");
			pwc_write(DA9052_CHGBUCK_REG, 0x0d, 0x0f);
		} else if (battery_data->charger == CHARGER_USB_PC100) {
			DEBUG_PRINT("change normalB, ISET_BUCK 100mA\n");
			low_speed_mode();
			pwc_write(DA9052_CHGBUCK_REG, 0x03, 0x0f);
		} else {
			DEBUG_PRINT("ISET_BUCK 500mA \n");
			pwc_write(DA9052_CHGBUCK_REG, 0x08, 0x0f);
		}

		if (old_charger == CHARGER_USB_PC100) {
			DEBUG_PRINT("change normalB\n");
			full_speed_mode();
		}
	}

	if (battery_data->charger == CHARGER_BATTERY) {
		if (battery_data->level < 10) {
			DEBUG_PRINT("Deep Sleep\n");
// eflow-removed
//			mp200_pm_sleep(PM_SLEEP_MODE_S3); /* DeepSleep */
		} else if (battery_data->level < 30) {
		#if 0
			low_speed_mode();
		#endif
		}
	}
}

/* set Battery charger voltage limit (CV) */
void mp200_battery_set_voltlim(int limit)
{
	if ((limit < VCHG_BAT_3_650V) || (VCHG_BAT_4_425V < limit))
		return;

	voltage_limit = limit;
// eflow-removed
//	pwc_write(DA9052_CHGCONT_REG, voltage_limit << 3, 0xf8);
}


/* set Battery charger current limit (CC) */
void mp200_battery_set_curlim(int limit)
{
	if ((limit < ICHG_BAT_0MA) || (ICHG_BAT_1260MA < limit))
		return;

	current_limit = limit;
// eflow-removed
//	pwc_write(DA9052_BATCHG_REG, current_limit, 0x3f);
}



static irqreturn_t mp200_charger_irq(int irq, void *dev_id)
{
	if (irq == INT_PWC_E_CHG_END) {
		printk(KERN_INFO "Charge completion\n");
	} else {
#ifdef MP200_BATTERY_DEBUG
		if (irq == INT_PWC_E_DCIN_DET)
			printk(KERN_INFO "INT_PWC_E_DCIN_DET\n");
		else if (irq == INT_PWC_E_DCIN_REM)
			printk(KERN_INFO "INT_PWC_E_DCIN_REM\n");
		else if (irq == INT_PWC_E_VBUS_DET)
			printk(KERN_INFO "INT_PWC_E_VBUS_DET\n");
		else if (irq == INT_PWC_E_VBUS_REM)
			printk(KERN_INFO "INT_PWC_E_VBUS_REM\n");
#endif
		queue_work(mp200_battery_workqueue, &mp200_battery_work);
	}

	return IRQ_HANDLED;
}

static int mp200_battery_get_status(void)
{
	int ret;

	switch (battery_data->charger) {
	case CHARGER_BATTERY:
		ret = POWER_SUPPLY_STATUS_NOT_CHARGING;
		break;
	case CHARGER_USB_CG:
	case CHARGER_USB_PC100:
	case CHARGER_USB_PC500:
	case CHARGER_AC:
		if (battery_data->level == 100)
			ret = POWER_SUPPLY_STATUS_FULL;
		else
			ret = POWER_SUPPLY_STATUS_CHARGING;
		break;
	default:
		ret = POWER_SUPPLY_STATUS_UNKNOWN;
	}
	return ret;
}

static int mp200_battery_get_property(struct power_supply *psy,
				 enum power_supply_property psp,
				 union power_supply_propval *val)
{
	int ret = 0;

	switch (psp) {
	case POWER_SUPPLY_PROP_STATUS:	/* 0 */
		val->intval = mp200_battery_get_status();
		break;
	case POWER_SUPPLY_PROP_HEALTH:	/* 1 */
		val->intval = POWER_SUPPLY_HEALTH_GOOD;
		break;
	case POWER_SUPPLY_PROP_PRESENT:	/* 2 */
		val->intval = (battery_data->battery_det ==
					 CHARGER_BATTERY ? 1 : 0);
		break;
	case POWER_SUPPLY_PROP_TECHNOLOGY:	/* 4 */
		val->intval = POWER_SUPPLY_TECHNOLOGY_LION;
		break;
	case POWER_SUPPLY_PROP_CAPACITY:	/* 26 */
		val->intval = battery_data->level;
		break;
// BEGIN eflow-added
	case POWER_SUPPLY_PROP_VOLTAGE_NOW:	/* 9 */
		val->intval = battery_data->voltage_now;
		break;
	case POWER_SUPPLY_PROP_CHARGE_NOW:	/* 17 */
		val->intval = battery_data->charge_now;
		break;
// END eflow-added
	default:
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int mp200_power_get_property(struct power_supply *psy,
			enum power_supply_property psp,
			union power_supply_propval *val)
{
	enum charger_type_t charger;

	charger = battery_data->charger;

	switch (psp) {
	case POWER_SUPPLY_PROP_ONLINE:	/* 3 */
		if (psy->type == POWER_SUPPLY_TYPE_MAINS) {
			val->intval = (charger ==  CHARGER_AC ? 1 : 0);
		} else if (psy->type == POWER_SUPPLY_TYPE_USB) {
			if ((charger == CHARGER_USB_CG) ||
				(charger == CHARGER_USB_PC100) ||
				(charger == CHARGER_USB_PC500)) {
				val->intval = 1;
			} else {
				val->intval = 0;
			}
		} else {
			val->intval = 0;
		}
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static void mp200_battery_watch_timer_func(unsigned long data)
{
	mp200_battery_watch();

	if (battery_data->charger == CHARGER_BATTERY) {
		if (battery_data->level < 10)
			queue_work(mp200_battery_workqueue,
							&mp200_battery_work);
		else if (battery_data->level < 30) {
// eflow-removed
//			printk(KERN_INFO "Please charge battery.\n");
		#if 0
			queue_work(mp200_battery_workqueue,
							&mp200_battery_work);
		#endif
		}
	} else if ((battery_data->battery_det == CHARGER_BATTERY) &&
				(battery_data->charger == CHARGER_USB_PC100)) {
		queue_work(mp200_battery_workqueue, &mp200_battery_work);
	}

	mod_timer(&mp200_battery_watch_timer,
		jiffies + msecs_to_jiffies(MP200_CHARGE_CHGFALL_WATCH_TIME));
}


static int mp200_battery_probe(struct platform_device *pdev)
{
	int ret;
	struct mp200_battery_data *data;

	data = kzalloc(sizeof(*data), GFP_KERNEL);
	if (data == NULL) {
		ret = -ENOMEM;
		goto err_data_alloc_failed;
	}
	spin_lock_init(&data->lock);

	/* Thermal charging control disabled */
	pwc_write(DA9052_CHGBUCK_REG, 0x00, 0x80);

	/* Battery pre-charge current limit=40mA */
	pwc_write(DA9052_BATCHG_REG, 0x80, 0xC0);

// eflow-added
	pwc_write(DA9052_ADCCONT_REG, 0x01, 0x01);

	 /* VCHG_BAT */
	mp200_battery_set_voltlim(CHARGING_VOLTAGE_LIMIT);

	/* ICHG_BAT */
	mp200_battery_set_curlim(CHARGING_CURRENT_LIMIT);

	mp200_battery_workqueue =
			create_singlethread_workqueue(MP200_BATTERY_NAME);
	INIT_WORK(&mp200_battery_work, mp200_charger_update_status);

	/* Battey */
	data->battery.name = "battery";
	data->battery.type = POWER_SUPPLY_TYPE_BATTERY;
	data->battery.properties = mp200_battery_props;
	data->battery.num_properties = ARRAY_SIZE(mp200_battery_props);
	data->battery.get_property = mp200_battery_get_property;

	/* USB */
	data->usb.name = "usb";
	data->usb.type = POWER_SUPPLY_TYPE_USB;
	data->usb.properties = mp200_power_props;
	data->usb.num_properties = ARRAY_SIZE(mp200_power_props);
	data->usb.get_property = mp200_power_get_property;

	/* AC */
	data->ac.name = "ac";
	data->ac.type = POWER_SUPPLY_TYPE_MAINS;
	data->ac.properties = mp200_power_props;
	data->ac.num_properties = ARRAY_SIZE(mp200_power_props);
	data->ac.get_property = mp200_power_get_property;

	battery_data = data;

	mp200_charger_get_status();

#if 0 // BEGIN eflow-removed
	ret = request_irq(INT_PWC_E_DCIN_DET, mp200_charger_irq,
			IRQF_SHARED, "charger", data);
	if (ret)
		goto err_irq_ac1;

	ret = request_irq(INT_PWC_E_DCIN_REM, mp200_charger_irq,
			IRQF_SHARED, "charger", data);
	if (ret)
		goto err_irq_ac2;
#endif // END eflow-removed

#if 0 // BEGIN eflow-removed
	ret = request_irq(INT_PWC_E_VBUS_DET, mp200_charger_irq,
			IRQF_SHARED, "charger", data);
	if (ret)
		goto err_irq_usb1;

	ret = request_irq(INT_PWC_E_VBUS_REM, mp200_charger_irq,
			IRQF_SHARED, "charger", data);
	if (ret)
		goto err_irq_usb2;
#endif // END eflow-removed

	ret = request_irq(INT_PWC_E_CHG_END, mp200_charger_irq,
			IRQF_SHARED, "charger", data);
	if (ret)
		goto err_irq_chg_end;

	ret = power_supply_register(&pdev->dev, &data->battery);
	if (ret)
		goto err_battery_failed;

	ret = power_supply_register(&pdev->dev, &data->usb);
	if (ret)
		goto err_usb_failed;

	ret = power_supply_register(&pdev->dev, &data->ac);
	if (ret)
		goto err_ac_failed;

	platform_set_drvdata(pdev, data);

	/* battery watch */
	init_timer(&mp200_battery_watch_timer);
	setup_timer(&mp200_battery_watch_timer,
			mp200_battery_watch_timer_func, 0);
	mod_timer(&mp200_battery_watch_timer,
		jiffies + msecs_to_jiffies(MP200_CHARGE_CHGFALL_WATCH_TIME));

	printk(KERN_INFO "Starting battery\n");

	return 0;

err_ac_failed:
	power_supply_unregister(&data->usb);
err_usb_failed:
	power_supply_unregister(&data->battery);
err_battery_failed:
	free_irq(INT_PWC_E_CHG_END, data);
err_irq_chg_end:
	free_irq(INT_PWC_E_VBUS_REM, data);
err_irq_usb2:
	free_irq(INT_PWC_E_VBUS_DET, data);
err_irq_usb1:
	free_irq(INT_PWC_E_DCIN_REM, data);
err_irq_ac2:
	free_irq(INT_PWC_E_DCIN_DET, data);
err_irq_ac1:
	kfree(data);
err_data_alloc_failed:
	return ret;
}

static int mp200_battery_remove(struct platform_device *pdev)
{
	struct mp200_battery_data *data = platform_get_drvdata(pdev);

	del_timer(&mp200_battery_watch_timer);

	free_irq(INT_PWC_E_CHG_END, data);
	free_irq(INT_PWC_E_DCIN_DET, data);
	free_irq(INT_PWC_E_DCIN_REM, data);
	free_irq(INT_PWC_E_VBUS_DET, data);
	free_irq(INT_PWC_E_VBUS_REM, data);

	power_supply_unregister(&data->battery);
	power_supply_unregister(&data->usb);
	power_supply_unregister(&data->ac);

	destroy_workqueue(mp200_battery_workqueue);

	kfree(data);
	battery_data = NULL;
	return 0;
}

static struct platform_driver mp200_battery_driver = {
	.probe		= mp200_battery_probe,
	.remove		= mp200_battery_remove,
	.driver = {
		.name = MP200_BATTERY_NAME
	}
};

static struct platform_device *mp200_battery_device;

static int __init mp200_battery_init(void)
{
	return platform_driver_register(&mp200_battery_driver);
}

static void __exit mp200_battery_exit(void)
{
	platform_device_unregister(mp200_battery_device);
	platform_driver_unregister(&mp200_battery_driver);
}

module_init(mp200_battery_init);
module_exit(mp200_battery_exit);

MODULE_AUTHOR("NECEL");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Battery driver for the mp200");
