/*
 * EHCI HCD (Host Controller Driver) for MP200 High speed USB Host controller.
 *
 *  Copyright (C) NEC Electronics Corporation 2009
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/platform_device.h>

#include <mach/pmu.h>
#include <mach/smu.h>
#include <mach/pwc.h>
#include <mach/gpio.h>
#include <mach/pm.h>


/*----------------------------------*/
/* Debug message */
/*#define USE_DEBUG_MESSAGE_ALL  */
/*#define USE_DEBUG_MESSAGE      */
/*#define USE_INFOMATION_MESSAGE */

#include "ehci-mp200.h"

/*---------------------------------------------------------------------------*/
/* */
static const char driver_name[] = "mp200-ehci-driver";

#ifdef USE_PHY_TYPE
static T_PHY_TYPE g_phy_type[] = {
	 /*  PHY      VID     PID  */
	{UNKNOWN_PHY, 0x0000, 0x0000},
	{    NXP_PHY, 0x04cc, 0x0000},		/* NXP */
	{   SMSC_PHY, 0x0424, 0x0000},		/* SMSC */
};
#endif	/* USE_PHY_TYPE */

/* ----- Work ----- */
struct mp200_hcd_work	*pg_hcd_work;

/*---------------------------------------------------------------------------*/
/* Function */
static int _mp200_init(struct platform_device *pdev);
static int _mp200_exit(struct platform_device *pdev);
static void _mp200_vbus_control(struct usb_hcd *hcd, int on_flag);
static void _mp200_enable_controller(struct usb_hcd *hcd);
static void _mp200_disable_controller(void);

#ifdef USE_DUMP_REGISTER
/*-------------------------------------------------------------------------*/
/*
 * PHY Register Dump
*/
static void _mp200_dump_phy_reg(struct usb_hcd *hcd)
{
	int		ndata;
	PT_OTG_REGS	p_regs = (PT_OTG_REGS)hcd->regs;

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT, 0x00);
	if (ndata < 0) {
		ERR_MSG("***** ULPI Read Error!!");
		return;
	}
	INFO_MSG("MP200-EHCI(PHY) : Vendor ID Low = 0x%02x", ndata);

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT, 0x01);
	INFO_MSG("MP200-EHCI(PHY) : Vendor ID High = 0x%02x", ndata);

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT, 0x02);
	INFO_MSG("MP200-EHCI(PHY) : Product ID Low = 0x%02x", ndata);

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT, 0x03);
	INFO_MSG("MP200-EHCI(PHY) : Product ID High = 0x%02x", ndata);

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT, 0x04);
	INFO_MSG("MP200-EHCI(PHY) : Function Control = 0x%02x", ndata);

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT, 0x0A);
	INFO_MSG("MP200-EHCI(PHY) : OTG Control = 0x%02x", ndata);

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT, 0x0D);
	INFO_MSG("MP200-EHCI(PHY) : USB Interrupt Enable = 0x%02x", ndata);
}

/*-------------------------------------------------------------------------*/
/*
 * Link Register Dump
*/
static void _mp200_dump_controller_reg(struct usb_hcd *hcd)
{
	int		i;
	u32 reg_data;

	for (i = 0x100 ; i < 0x200 ; i += 16) {
		reg_data = mp200st_udc_readl(
			(u32 *)IO_ADDRESS(MP200_USB_BASE + i));
		printk(KERN_INFO "MP200-EHCI(HC) : USB%04x = %08x", i,
			(int)reg_data);

		reg_data = mp200st_udc_readl(
			(u32 *)IO_ADDRESS(MP200_USB_BASE + i+4));
		printk(KERN_INFO " %08x", (int)reg_data);

		reg_data = mp200st_udc_readl(
			(u32 *)IO_ADDRESS(MP200_USB_BASE + i+8));
		printk(KERN_INFO " %08x", (int)reg_data);

		reg_data = mp200st_udc_readl(
			(u32 *)IO_ADDRESS(MP200_USB_BASE + i+12));
		printk(KERN_INFO " %08x\n", (int)reg_data);
	}
}
#endif	/* USE_DUMP_REGISTER */

/*-------------------------------------------------------------------------*/
/*
 * initialize
 *   initialize for Work area
*/
static int _mp200_init(struct platform_device *pdev)
{
	int	nret = 0;
	struct mp200_hcd_work	*p_work;

	pg_hcd_work = NULL;	/* global */

	p_work = kzalloc(sizeof(struct mp200_hcd_work), GFP_KERNEL);
	if (p_work != NULL) {
		pg_hcd_work = p_work;

		spin_lock_init(&p_work->lock);
		p_work->dma_mask = DMA_32BIT_MASK;

		/*-----------------------------------------------------------*/
		/* DMA parameter setting */
		pdev->dev.dma_mask = &p_work->dma_mask;
		pdev->dev.coherent_dma_mask = DMA_32BIT_MASK;

#ifdef USE_PHY_TYPE
		p_work->phy_type = UNKNOWN_PHY;
#endif	/* USE_PHY_TYPE */
	} else {
		nret = -ENOMEM;
	}

	return nret;
}

/*-------------------------------------------------------------------------*/
/*
 * end
 *   Work area memory free
*/
static int _mp200_exit(struct platform_device *pdev)
{
	int		nret = 0;

	if (pg_hcd_work != NULL) {
		kfree(pg_hcd_work);
		pg_hcd_work = NULL;
	}

	return nret;
}

#ifdef USE_PHY_TYPE
/*-------------------------------------------------------------------------*/
/*
 * get PHY ID
*/
static u16 _mp200_get_phy_id(struct usb_hcd *hcd)
{
	int		ndata;
	__u16		phyid;
	PT_OTG_REGS		p_regs = (PT_OTG_REGS)hcd->regs;

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT,
			PHY_VENDOR_ID_HIGH);
	if (ndata == -EIO)
		return 0;

	phyid = (u16)((ndata & 0xFF) << 8);

	ndata = mp200st_udc_ulpi_read(&p_regs->ULPIVIEWPORT, PHY_VENDOR_ID_LOW);
	if (ndata == -EIO)
		return 0;

	phyid |= (u16)(ndata & 0xFF);

	return phyid;
}

/*-------------------------------------------------------------------------*/
/*
 * get PHY Type
*/
static int _mp200_get_phy_type(
	struct mp200_hcd_work *p_work,
	struct usb_hcd *hcd)
{
	int	i, nret = -1;
	__u16	phyid;
	PT_PHY_TYPE	p_phy_type = g_phy_type;

	phyid = _mp200_get_phy_id(hcd);
	INFO_MSG("------ PHY Vendor ID = 0x%04x", phyid);

	for (i = 0; i < END_PHY; i++) {
		if (phyid == p_phy_type->vendorID) {
			p_work->phy_type = p_phy_type->phy_type;
			nret = 0;
			break;
		}
		p_phy_type++;
	}

	return nret;
}
#endif	/* USE_PHY_TYPE */


/*-------------------------------------------------------------------------*/
/*
 * VBUS ON/OFF
 *
 * on_flag
 *    0    : VBUS OFF
 *  !=0    : VBUS ON
*/
static void _mp200_vbus_control(struct usb_hcd *hcd, int on_flag)
{
	PT_OTG_REGS	p_regs = (PT_OTG_REGS)hcd->regs;
	__u8		data;


	if (system_rev & 0x40) {
#ifndef CONFIG_MP200_SDMOBILE
		if (on_flag == 0)
			gpio_direction_output(GPIO_CN9_1PIN, 1);
		else
			gpio_direction_output(GPIO_CN9_1PIN, 0);
#endif
	} else {
#ifdef USE_PHY_TYPE
		switch (pg_hcd_work->phy_type) {
		case SMSC_PHY:
			data = (BIT06 | BIT05);
			break;

		case NXP_PHY:
		default:
			data = BIT06;
			break;
		}
#else	/* !USE_PHY_TYPE */
		data = BIT06;
#endif	/* USE_PHY_TYPE */

		if (on_flag == 0) {
			DBG_MSG(" --- VBUS OFF");
			mp200st_udc_ulpi_write(&p_regs->ULPIVIEWPORT,
				PHY_OTG_CONTROL_C, data);
		} else {
			DBG_MSG(" --- VBUS ON");
			mp200st_udc_ulpi_write(&p_regs->ULPIVIEWPORT,
				PHY_OTG_CONTROL_S, data);
		}
	}
}

/*-------------------------------------------------------------------------*/
/*
 * USB Controller Enable
*/
static void _mp200_enable_controller(struct usb_hcd *hcd)
{
	PT_OTG_REGS	p_regs = (PT_OTG_REGS)hcd->regs;

	INFO_MSG("mp200_enable_controller() Enter");

	/*---------------------------------------------------------------*/
	/* PHY */
#ifndef	CONFIG_MP200_EM1_DKIT
	pwc_write(PWC_PSCNT1, PHY_POWER_BIT, PHY_POWER_BIT);	/* Power ON */
#endif
	if (system_rev & 0x40)
		gpio_direction_output(UDC_PHY_CS, 1);		/* RESETB */
	else
		gpio_direction_output(UDC_PHY_CS, 0);		/* CS */

	mdelay(UDC_PHY_CS_MDELAY);

	/*---------------------------------------------------------------*/
	/* Link */
	/* CLK supply ON */
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_USB_CLK);
	/* Reset OFF */
	mp200_pmu_unreset_device(MP200_RESETDEVICE_USB);

	udelay(UDC_LINK_UNRESET_UDELAY);

	/*---------------------------------------------------------------*/
	/* set to Host */
	/* Host Mode */
	mp200st_udc_set_bit(&p_regs->USB_MODE, UDBMODE_CM_HC);

#ifdef USE_PHY_TYPE
	_mp200_get_phy_type(pg_hcd_work, hcd);
#endif	/* USE_PHY_TYPE */

}

/*-------------------------------------------------------------------------*/
/*
 * USB Controller Disable
*/
static void _mp200_disable_controller(void)
{
	INFO_MSG("mp200_disable_controller() Enter");

	/*---------------------------------------------------------------*/
	/* Link disable */
	/* Reset ON */
	mp200_pmu_reset_device(MP200_RESETDEVICE_USB);
	/* CLK supply OFF */
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_USB_CLK);

	/*---------------------------------------------------------------*/
	/* PHY disable */

	if (system_rev & 0x40)
		gpio_direction_output(UDC_PHY_CS, 0);	/* RESETB */
	else
		gpio_direction_output(UDC_PHY_CS, 1);	/* CS */

#ifndef	CONFIG_MP200_EM1_DKIT
	/* Power OFF */
	pwc_write(PWC_PSCNT1, 0, PHY_POWER_BIT);
#endif
}

/*-------------------------------------------------------------------------*/
/*
 * USB Controller Reset
*/
static int mp200_ehci_hc_reset(struct usb_hcd *hcd)
{
	struct ehci_hcd *ehci = hcd_to_ehci(hcd);
	int		retval;

	INFO_MSG("mp200_ehci_hc_reset() Enter");

	if (ehci == NULL) {
		ERR_MSG("ehci == NULL !!");
		return -EFAULT;
	}

	/*---------------------------------------------------------------*/
	/* set USB Controller enable, and set any paramters */
	_mp200_enable_controller(hcd);

	ehci->caps = hcd->regs + MP200_EHCI_CAPLENGTH_OFFSET;
	ehci->regs = hcd->regs + MP200_EHCI_CAPLENGTH_OFFSET +
			HC_LENGTH(ehci_readl(ehci, &ehci->caps->hc_capbase));
	ehci->hcs_params = ehci_readl(ehci, &ehci->caps->hcs_params);
	ehci->sbrn = 0x20;
	hcd->has_tt = 1;

	/*---------------------------------------------------------------*/
	/* stop EHCI */
	retval = ehci_halt(ehci);
	if (retval != 0) {
		ERR_MSG("ehci_halt() !!");
		return retval;
	}

	/*---------------------------------------------------------------*/
	/* initialize EHCI */
	retval = ehci_init(hcd);
	if (retval != 0) {
		ERR_MSG("ehci_init() !!");
		return retval;
	}

	/*---------------------------------------------------------------*/
	/* Port Power OFF */
	ehci_port_power(ehci, 0);

	return retval;
}


/*-------------------------------------------------------------------------*/
/*
 * USB Controller Start(Run)
*/
static int mp200_ehci_hc_start(struct usb_hcd *hcd)
{
	int		retval;
	PT_OTG_REGS	p_regs = (PT_OTG_REGS)hcd->regs;

	INFO_MSG("mp200_ehci_hc_start() Enter");

	retval = ehci_run(hcd);

	/*---------------------------------------------------------------*/
	/* set Host mode                                            */
	/*   call reset in ehci_run() function. that is reset Link. */
	/*   this Link is disable host mode by reset, */
	/*   need re-setting. */
	mp200st_udc_set_bit(&p_regs->USB_MODE, UDBMODE_CM_HC);	/* Host Mode */

#ifdef USE_DUMP_REGISTER
	_mp200_dump_phy_reg(hcd);
	_mp200_dump_controller_reg(hcd);
#endif	/* USE_DUMP_REGISTER */

	return retval;
}

/*-------------------------------------------------------------------------*/
/*
 * Root HUB Control
*/
static int mp200_ehci_hub_control(
	struct usb_hcd	*hcd,
	u16		typeReq,
	u16		wValue,
	u16		wIndex,
	char	*buf,
	u16		wLength
)
{
	int		retval;

	INFO_MSG("mp200_ehci_hub_control() Enter");

	retval = ehci_hub_control(hcd, typeReq, wValue, wIndex, buf, wLength);

	/*---------------------------------------------------------------*/
	/* VBUS ON/OFF control */
	if (retval == 0) {
		if (wValue == USB_PORT_FEAT_POWER) {
			if (typeReq == SetPortFeature)
				_mp200_vbus_control(hcd, 1);
			else if (typeReq == ClearPortFeature)
				_mp200_vbus_control(hcd, 0);

		}
	}

	return retval;
}

/*-------------------------------------------------------------------------*/

static const struct hc_driver ehci_mp200_hc_driver = {
	.description = hcd_name,
	.product_desc = "MP200 EHCI Host Controller",
	.hcd_priv_size = sizeof(struct ehci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq = ehci_irq,
	.flags = HCD_MEMORY | HCD_USB2,

	/*
	 * basic lifecycle operations
	 */
	.reset = mp200_ehci_hc_reset,
	.start = mp200_ehci_hc_start,
	.stop = ehci_stop,
	.shutdown = ehci_shutdown,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue = ehci_urb_enqueue,
	.urb_dequeue = ehci_urb_dequeue,
	.endpoint_disable = ehci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number = ehci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data = ehci_hub_status_data,
	.hub_control = mp200_ehci_hub_control,

#ifdef	CONFIG_PM
	.bus_suspend = ehci_bus_suspend,
	.bus_resume = ehci_bus_resume,
#endif

	.relinquish_port = ehci_relinquish_port,
	.port_handed_over = ehci_port_handed_over,

};

/*-------------------------------------------------------------------------*/
/*
 * Probe
*/
static int mp200_ehci_pd_probe(struct platform_device *pdev)
{
	struct usb_hcd *hcd = NULL;
	const struct hc_driver *driver = &ehci_mp200_hc_driver;
	int		retval = 0;

	dev_dbg(&pdev->dev, "mp200_ehci_pd_probe() Enter\n");

	/*---------------------------------------------------------------*/
	if (usb_disabled()) {
		dev_err(&pdev->dev, "***** USB Dissabled !!\n");
		return -ENODEV;
	}

	/*---------------------------------------------------------------*/
	/* get Work */
	retval = _mp200_init(pdev);
	if (retval != 0) {
		dev_err(&pdev->dev, "***** _mp200_init() Error !!\n");
		return retval;
	}

	/*---------------------------------------------------------------*/
	/* make HCD memory */
	hcd = usb_create_hcd(driver, &pdev->dev, pdev->dev.bus_id);
	if (hcd == NULL) {
		dev_err(&pdev->dev, "***** usb_create_hcd() Error !!\n");
		retval = -ENOMEM;
		goto err_create_hcd;
	}

	/*---------------------------------------------------------------*/
	/* setting HCD parameters */
	hcd->rsrc_start = MP200_USB_BASE;
	hcd->rsrc_len = MP200_USB_SIZE;

	DBG_MSG(" --- hcd->rsrc_start = 0x%08x", (u32)hcd->rsrc_start);
	DBG_MSG(" --- hcd->rsrc_len   = 0x%08x", (u32)hcd->rsrc_len);

	/*---------------------------------------------------------------*/
	/* HCD Register */
	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, driver_name)) {
		dev_err(&pdev->dev, "***** request_mem_region() Error !!\n");
		retval = -EBUSY;
		goto err_mem_region;
	}

	/*---------------------------------------------------------------*/
	/* memory map for HCD Register */
	hcd->regs = ioremap(hcd->rsrc_start, hcd->rsrc_len);
	if (hcd->regs == NULL) {
		dev_err(&pdev->dev, "***** ioremap() Error !!\n");
		retval = -EFAULT;
		goto err_ioremap;
	}

	/*---------------------------------------------------------------*/
	/* add HCD */
	retval = usb_add_hcd(hcd, UDC_USB_INT, IRQF_SHARED|IRQF_DISABLED);
	if (retval != 0) {
		dev_err(&pdev->dev, "***** usb_add_hcd() Error !!\n");
		goto err_add_hcd;
	}

	return retval;

err_add_hcd:
	iounmap(hcd->regs);

err_ioremap:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);

err_mem_region:
	usb_put_hcd(hcd);

err_create_hcd:
	_mp200_exit(pdev);

	dev_err(&pdev->dev, "***** mp200_ehci_pd_probe(%d) Retern\n", retval);

	return retval;
}

/*-------------------------------------------------------------------------*/
/*
 * Remove
*/
static int mp200_ehci_pd_remove(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	dev_dbg(&pdev->dev, "mp200_ehci_pd_remove() Enter\n");

	/*---------------------------------------------------------------*/
	/* remove HDC */
	usb_remove_hcd(hcd);

	/*---------------------------------------------------------------*/
	/* disable memory map */
	iounmap(hcd->regs);

	/*---------------------------------------------------------------*/
	/* release memory */
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);

	/*---------------------------------------------------------------*/
	/* release HCD */
	usb_put_hcd(hcd);

	/*---------------------------------------------------------------*/
	/* disable HCD */
	_mp200_disable_controller();

	/*---------------------------------------------------------------*/
	/* free Work area */
	_mp200_exit(pdev);

	return 0;
}

/*-------------------------------------------------------------------------*/
/*
 * Suspend
*/
static int mp200_ehci_pd_suspend(
	struct platform_device *pdev,
	pm_message_t message)
{
#if 0
	struct usb_hcd *hcd = platform_get_drvdata(pdev);
	struct ehci_hcd *ehci = hcd_to_ehci(hcd);
	unsigned long flags;
	int rc = 0;

	dev_dbg(&pdev->dev, "mp200_ehci_pd_suspend() Enter\n");

	if (time_before(jiffies, ehci->next_statechange))
		msleep(10);

	spin_lock_irqsave(&ehci->lock, flags);
	if (hcd->state != HC_STATE_SUSPENDED) {
		rc = -EINVAL;
		goto bail;
	}
	ehci_writel(ehci, 0, &ehci->regs->intr_enable);
	(void)ehci_readl(ehci, &ehci->regs->intr_enable);

	if (message.event == PM_EVENT_PRETHAW) {
		ehci_halt(ehci);
		ehci_reset(ehci);
	}

	clear_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

	_mp200_disable_controller();

 bail:
	spin_unlock_irqrestore(&ehci->lock, flags);

	/* could save FLADJ in case of Vaux power loss */
	/* ... we'd only use it to handle clock skew   */

	return rc;
#else
	return -EBUSY;
#endif
}

/*-------------------------------------------------------------------------*/
/*
 * Resume
*/
static int mp200_ehci_pd_resume(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);
	struct ehci_hcd *ehci = hcd_to_ehci(hcd);

	dev_dbg(&pdev->dev, "mp200_ehci_pd_resume() Enter\n");

	_mp200_enable_controller(hcd);

	/* maybe restore FLADJ */

	if (time_before(jiffies, ehci->next_statechange))
		msleep(100);

	/* Mark hardware accessible again as we are out of D3 state by now */
	set_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

	/* If CF is still set, we maintained PCI Vaux power.
	 * Just undo the effect of ehci_pci_suspend().
	 */
	if (ehci_readl(ehci, &ehci->regs->configured_flag) == FLAG_CF) {
		int	mask = INTR_MASK;

		if (!hcd->self.root_hub->do_remote_wakeup)
			mask &= ~STS_PCD;
		ehci_writel(ehci, mask, &ehci->regs->intr_enable);
		ehci_readl(ehci, &ehci->regs->intr_enable);
		return 0;
	}

	ehci_dbg(ehci, "lost power, restarting\n");
	usb_root_hub_lost_power(hcd->self.root_hub);

	/* Else reset, to cope with power loss or flush-to-storage
	 * style "resume" having let BIOS kick in during reboot.
	 */
	(void) ehci_halt(ehci);
	(void) ehci_reset(ehci);

	/* emptying the schedule aborts any urbs */
	spin_lock_irq(&ehci->lock);
	if (ehci->reclaim)
		end_unlink_async(ehci);
	ehci_work(ehci);
	spin_unlock_irq(&ehci->lock);

	ehci_writel(ehci, ehci->command, &ehci->regs->command);
	ehci_writel(ehci, FLAG_CF, &ehci->regs->configured_flag);
	ehci_readl(ehci, &ehci->regs->command);	/* unblock posted writes */

	/* here we "know" root ports should always stay powered */
	ehci_port_power(ehci, 1);

	hcd->state = HC_STATE_SUSPENDED;

	return 0;
}


MODULE_ALIAS("mp200-ehci");


static struct platform_driver ehci_mp200_driver = {
	.probe    = mp200_ehci_pd_probe,
	.remove   = mp200_ehci_pd_remove,
	.shutdown = usb_hcd_platform_shutdown,
	.suspend  = mp200_ehci_pd_suspend,
	.resume   = mp200_ehci_pd_resume,
	.driver = {
		.owner = THIS_MODULE,
		.name  = (char *) driver_name,
	}
};

