/*
 * File Name       : /drivers/video/mp200/mp200_imc.c
 * Function        : IMC driver corresponding to V4L2
 * Release Version :
 * Release Date    : 2009.10.21
 *
 * Copyright (C) NEC Electronics Corporation 2009
 *
 *
 * This program is free software;you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by Free Softwere
 * Foundation; either version 2 of License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warrnty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; If not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/********************************************************
 *  Include Files                                       *
 *******************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/platform_device.h>
#include <linux/io.h>

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
#include <mach/pm.h>
#endif /* CONFIG_PM || CONFIG_DPM */

#include <mach/hardware.h>
#include <mach/irqs.h>
#include <mach/pmu.h>
#include <mach/em1_mem.h>

#include "imc.h"
#include "mp200_imc.h"
#include "mp200_common.h"


/********************************************************
 *  Definitions                                         *
 *******************************************************/
#define _DEBUG_IMC  0x00 /* 00008421(bit) */
			 /* 0x01: debug function in
			  * 0x02: debug function out
			  */

#define LCD_ASYNCHRONOUS 0 /* 0: LCD-IMC synchronous mode
			    * 1: LCD-IMC asynchronous mode
			    */


/********************************************************
 *  Definitions                                         *
 *******************************************************/
#define DEV_NAME "mp200_imc"
#define DPM_SUSPEND_FLG_INIT 0
#define CTRL_FUNC_INIT 0


/********************************************************
 *  Macros                                              *
 *******************************************************/
#define printk_err(fmt, arg...) \
	do {                     \
		printk(KERN_ERR DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)

#define printk_info(fmt, arg...) \
	do {                      \
		printk(KERN_INFO DEV_NAME ": " fmt, ## arg); \
	} while (0)

#if _DEBUG_IMC
#define printk_dbg(level, fmt, arg...) \
	do {                            \
		if (level > 0) \
			printk(KERN_DEBUG DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)
#else
#define printk_dbg(level, fmt, arg...) \
	;
#endif

/********************************************************
 *  Structure                                           *
 *******************************************************/
struct mp200_imc_device {
	unsigned long             IMCMmio;
	unsigned long             IMCMmioLength;
	char                     *IMCMmioV;

	spinlock_t                imc_hw_lock;

	int                       chk_preset;

	imc_callback_func_refresh callback_func_refresh;
	imc_callback_func_wb      callback_func_wb;
};


/********************************************************
 *  Variables                                           *
 *******************************************************/
static struct mp200_imc_device *dev;


/********************************************************
 *  Prototype declarations of local function            *
 *******************************************************/
/* ----- internal function --------------------------- */
static int         imc_hw_init(void);
static void        imc_hw_int_enable(void);
static void        imc_hw_int_disable(void);
#if LCD_ASYNCHRONOUS /* this function checked to IMC-LCD asynchronous mode. */
static int         imc_hw_chk_startmode(void);
#endif
static int         imc_hw_chk_busy(void);
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
static int         imc_hw_chk_status(void);
#endif /* CONFIG_PM || CONFIG_DPM */

/* ----- IMC IRQ handler ----------------------------- */
static irqreturn_t imc_irq_handler(int irq, void *dev_id);
static void        imc_irq_handler_sub(int irq, void *dev_id);

/* ----- suspend/resume function ------------------------------------------- */
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
static void        imc_resume_kick(unsigned long num);

/* Tasklet */
static DECLARE_TASKLET(imc_nexttask, imc_resume_kick, 0);

/* Flags */
static int DPM_suspend_flg = DPM_SUSPEND_FLG_INIT;
static int ctrl_func       = CTRL_FUNC_INIT;

/* Macro */
#define Call_StatusCtrlFunc_ON()  \
	do { ctrl_func = 1; } while (0)
#define Call_StatusCtrlFunc_OFF() \
	do { ctrl_func = 0; } while (0)
#else
#define Call_StatusCtrlFunc_ON()
#define Call_StatusCtrlFunc_OFF()
#endif /* CONFIG_PM || CONFIG_DPM */


/* ----- external function ------------------------------------------------- */
/******************************************************************************
* MODULE   : mp200_imc_set_preset
* FUNCTION :
* RETURN   :   0: success
*          : -16: EBUSY
*          : -22: EINVAL
* NOTE     : none
* CREATE   : 2007.06.21 M.Haguro
* UPDATE   :
******************************************************************************/
int  mp200_imc_set_preset(struct mp200_imc_preset *imc_preset, int setmode)
{
	unsigned long flags;
	int    gamma_cnt = 0;

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	if (!imc_preset) {
		printk_err("error! imc_preset(NULL)\n");
		return -EINVAL;
	}

	if (setmode) {
		/* IMC STATUS STOP WAITING */
		while (readl(dev->IMCMmioV + IMC_STATUS) != IMC_STATUS_STOP)
			msleep(1);
	}

	/*>>>>>> interuppt disable */
	spin_lock_irqsave(&dev->imc_hw_lock, flags);

	/* Function Settings Register */
	if (imc_preset->imc_control) {
		/* IMC Control                  */
		writel(*imc_preset->imc_control, dev->IMCMmioV + IMC_CONTROL);
	}

	if (imc_preset->gamma) {
		unsigned long *data = imc_preset->gamma->data;
		/* Gamma Control Register */
		if (imc_preset->gamma->en == IMC_GAMMAEN_ON) {
			unsigned long ulRegVal;

			ulRegVal = readl(dev->IMCMmioV + IMC_CONTROL);

			/* IMC Control                  */
			writel(ulRegVal & ~IMC_CLKCNT_GAMMA_ON,
			 dev->IMCMmioV + IMC_CONTROL);
			/* Gamma Correction Control     */
			writel(IMC_GAMMAEN_OFF, dev->IMCMmioV + IMC_GAMMA_EN);

			/* Gamma Control Table */
			/* Gamma Control Table Address  */
			writel(imc_preset->gamma->adr,
			 dev->IMCMmioV + IMC_GAMMA_ADR);
			for (gamma_cnt = 0; gamma_cnt < 64; gamma_cnt++ , data++) {
				/* Gamma Control Table Data     */
				writel(*data, dev->IMCMmioV + IMC_GAMMA_DATA);
			}

			/* IMC Control                  */
			writel(ulRegVal, dev->IMCMmioV + IMC_CONTROL);
		}
		/* Gamma Correction Control     */
		writel(imc_preset->gamma->en, dev->IMCMmioV + IMC_GAMMA_EN);
	}

#if LCD_ASYNCHRONOUS /* this function checked to IMC-LCD asynchronous mode. */
	if (imc_preset->wb && imc_hw_chk_startmode()) {
#else
	if (imc_preset->wb) {
#endif
		unsigned long wb_size;

		wb_size = ((imc_preset->wb->vsize << IMC_WB_VSIZE_SFT)
		 | imc_preset->wb->hsize);

		/* Start Exclusive Use Immediately Setting Register */
		/* WB Display Area Address      */
		writel(imc_preset->wb->areaadr, dev->IMCMmioV + IMC_WB_AREADDR);
		/* WB Address Increment Value   */
		writel(imc_preset->wb->hoffset, dev->IMCMmioV + IMC_WB_HOFFSET);
		/* WB Format                    */
		writel(imc_preset->wb->format, dev->IMCMmioV + IMC_WB_FORMAT);
		/* WB Image Size                */
		writel(wb_size, dev->IMCMmioV + IMC_WB_SIZE);
	}

	if (imc_preset->yuv) {
		/* All Layer Setting Register */
		/* Y Gain Offset                */
		writel(imc_preset->yuv->ygain, dev->IMCMmioV + IMC_YGAINOFFSET);
		/* U Gain Offset                */
		writel(imc_preset->yuv->ugain, dev->IMCMmioV + IMC_UGAINOFFSET);
		/* V Gain Offset                */
		writel(imc_preset->yuv->vgain, dev->IMCMmioV + IMC_VGAINOFFSET);
		/* YUV to RGB Conversion method */
		writel(imc_preset->yuv->yuv2rgb, dev->IMCMmioV + IMC_YUV2RGB);

		if ((imc_preset->yuv->yuv2rgb & IMC_TRANS_MODE_BIT) == IMC_TRANS_MODE_CUSTOM1 || (imc_preset->yuv->yuv2rgb & IMC_TRANS_MODE_BIT) == IMC_TRANS_MODE_CUSTOM2) {
			/* IMC_YUV2RGB Register selected Custom factor */
			/* Custom factor (Coef R0)      */
			writel(imc_preset->yuv->coef_r[0],
			 dev->IMCMmioV + IMC_COEF_R0);
			/* Custom factor (Coef R1)      */
			writel(imc_preset->yuv->coef_r[1],
			 dev->IMCMmioV + IMC_COEF_R1);
			/* Custom factor (Coef R2)      */
			writel(imc_preset->yuv->coef_r[2],
			 dev->IMCMmioV + IMC_COEF_R2);
			/* Custom factor (Coef R3)      */
			writel(imc_preset->yuv->coef_r[3],
			 dev->IMCMmioV + IMC_COEF_R3);

			/* Custom factor (Coef G0)      */
			writel(imc_preset->yuv->coef_g[0],
			 dev->IMCMmioV + IMC_COEF_G0);
			/* Custom factor (Coef G1)      */
			writel(imc_preset->yuv->coef_g[1],
			 dev->IMCMmioV + IMC_COEF_G1);
			/* Custom factor (Coef G2)      */
			writel(imc_preset->yuv->coef_g[2],
			 dev->IMCMmioV + IMC_COEF_G2);
			/* Custom factor (Coef G3)      */
			writel(imc_preset->yuv->coef_g[3],
			 dev->IMCMmioV + IMC_COEF_G3);

			/* Custom factor (Coef B0)      */
			writel(imc_preset->yuv->coef_b[0],
			 dev->IMCMmioV + IMC_COEF_B0);
			/* Custom factor (Coef B1)      */
			writel(imc_preset->yuv->coef_b[1],
			 dev->IMCMmioV + IMC_COEF_B1);
			/* Custom factor (Coef B2)      */
			writel(imc_preset->yuv->coef_b[2],
			 dev->IMCMmioV + IMC_COEF_B2);
			/* Custom factor (Coef B3)      */
			writel(imc_preset->yuv->coef_b[3],
			 dev->IMCMmioV + IMC_COEF_B3);
		}
	}

	if (imc_preset->keycolor) {
		/* Colorkey Register(Layer0, Layer1X) */
		/* Layer0  Color Key            */
		writel(imc_preset->keycolor->l0_keycolor,
		 dev->IMCMmioV + IMC_L0_KEYCOLOR);
		/* Layer1x Color Key            */
		writel(imc_preset->keycolor->l1x_keycolor,
		 dev->IMCMmioV + IMC_L1X_KEYCOLOR);
		printk_dbg((_DEBUG_IMC & 0x04), "l0_keycolor(%08lx)\n",
		 imc_preset->keycolor->l0_keycolor);
		printk_dbg((_DEBUG_IMC & 0x04), "l1_keycolor(%08lx)\n",
		 imc_preset->keycolor->l1x_keycolor);
	}

	if (imc_preset->bytelane) {
		/* Bytelane Register(Layer2a, Layer2b) */
		/* Layer2A Bytelane             */
		writel(imc_preset->bytelane->l2a_bytelane,
		 dev->IMCMmioV + IMC_L2A_BYTELANE);
		/* Layer2B Bytelane             */
		writel(imc_preset->bytelane->l2b_bytelane,
		 dev->IMCMmioV + IMC_L2B_BYTELANE);
		printk_dbg((_DEBUG_IMC & 0x04), "l2a_bytelane(%08lx)\n",
		 imc_preset->bytelane->l2a_bytelane);
		printk_dbg((_DEBUG_IMC & 0x04), "l2b_bytelane(%08lx)\n",
		 imc_preset->bytelane->l2b_bytelane);
	}

	dev->chk_preset = 1;

	/*<<<<<< interuppt enable */
	spin_unlock_irqrestore(&dev->imc_hw_lock, flags);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
	return 0;
}


/******************************************************************************
* MODULE   : mp200_imc_set_update_vsync
* FUNCTION :
* RETURN   :   0: success
*          : -13: EACCESS
*          : -22: EINVAL
* NOTE     : none
* CREATE   : 2007.06.21 M.Haguro
* UPDATE   :
******************************************************************************/
int  mp200_imc_set_update_vsync(struct mp200_imc_update_vsync *imc_vsync)
{
	unsigned long flags;

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	if (!imc_vsync) {
		printk_err("error! imc_vsync(NULL)\n");
		return -EINVAL;
	}

	if (!dev->chk_preset) {
		printk_err("error! need to \"mp200_imc_preset()\"\n");
		return -EACCES;
	}

	/*>>>>>> interuppt disable */
	spin_lock_irqsave(&dev->imc_hw_lock, flags);

	/* Double Buffer Control                  */
	writel(imc_vsync->cpubufsel,	dev->IMCMmioV + IMC_CPUBUFSEL);
	/* Horizontal, Vertical Mirroring Control */
	writel(imc_vsync->mirror,	dev->IMCMmioV + IMC_MIRROR);

	/*<<<<<< interuppt enable */
	spin_unlock_irqrestore(&dev->imc_hw_lock, flags);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
	return 0;
}


/******************************************************************************
* MODULE   : mp200_imc_set_update_reserve
* FUNCTION :
* RETURN   :   0: success
*          : -13: EACCESS
*          : -16: EBUSY
*          : -22: EINVAL
* NOTE     : none
* CREATE   : 2007.06.21 M.Haguro
* UPDATE   :
******************************************************************************/
int  mp200_imc_set_update_reserve(struct mp200_imc_update_reserve *imc_reserve,
 imc_callback_func_refresh callback_refresh)
{
	unsigned long flags;

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	/* register REFRESH */
	writel(IMC_UPDATE_OFF,	dev->IMCMmioV + IMC_REFRESH);

	if (!imc_reserve) {
		printk_err("error! imc_reserve(NULL)\n");
		return -EINVAL;
	}

	if (!dev->chk_preset) {
		printk_err("error! need to \"mp200_imc_preset()\"\n");
		return -EACCES;
	}

	if (!callback_refresh) {
		printk_err("error! imc_callback_func_refresh is invalid address\n");
		return -EINVAL;
	}

	/*>>>>>> interuppt disable */
	spin_lock_irqsave(&dev->imc_hw_lock, flags);

	if (imc_reserve->lo) {
		/*** Layer0 Setting Register ***/
		/* Layer0 Control                  */
		writel(imc_reserve->lo->control,
		 dev->IMCMmioV + IMC_L0_CONTROL);
		/* Layer0 Format                   */
		writel(imc_reserve->lo->format, dev->IMCMmioV + IMC_L0_FORMAT);
		/* Layer0 Color Key Control        */
		writel(imc_reserve->lo->keyenable,
		 dev->IMCMmioV + IMC_L0_KEYENABLE);
		/* Layer0 Alpha                    */
		writel(imc_reserve->lo->alpha, dev->IMCMmioV + IMC_L0_ALPHA);
		/* Layer0 Resize                   */
		writel(imc_reserve->lo->resize, dev->IMCMmioV + IMC_L0_RESIZE);
		/* Layer0 Address increment value  */
		writel(imc_reserve->lo->offset, dev->IMCMmioV + IMC_L0_OFFSET);
		/* Layer0 Address                  */
		writel(imc_reserve->lo->frameadr,
		 dev->IMCMmioV + IMC_L0_FRAMEADR);
		/* Layer0 LCD Out Positon          */
		writel(imc_reserve->lo->position,
		 dev->IMCMmioV + IMC_L0_POSITION);
		/* Layer0 LCD Out Size             */
		writel(imc_reserve->lo->size, dev->IMCMmioV + IMC_L0_SIZE);
	}

	if (imc_reserve->l1a) {
		/*** Layer1a Setting Register ***/
		/* Layer1A Control                 */
		writel(imc_reserve->l1a->control,
		 dev->IMCMmioV + IMC_L1A_CONTROL);
		/* Layer1A Format                  */
		writel(imc_reserve->l1a->format,
		 dev->IMCMmioV + IMC_L1X_FORMAT);
		/* Layer1A Color Key Control       */
		writel(imc_reserve->l1a->keyenable,
		 dev->IMCMmioV + IMC_L1A_KEYENABLE);
		/* Layer1A Alpha                   */
		writel(imc_reserve->l1a->alpha, dev->IMCMmioV + IMC_L1A_ALPHA);
		/* Layer1A Resize                  */
		writel(imc_reserve->l1a->resize,
		 dev->IMCMmioV + IMC_L1X_RESIZE);
		/* Layer1A Address increment value */
		writel(imc_reserve->l1a->offset,
		 dev->IMCMmioV + IMC_L1X_OFFSET);
		/* Layer1A Address                 */
		writel(imc_reserve->l1a->frameadr,
		 dev->IMCMmioV + IMC_L1A_FRAMEADR);
		/* Layer1A LCD Out Positon         */
		writel(imc_reserve->l1a->position,
		 dev->IMCMmioV + IMC_L1A_POSITION);
		/* Layer1A LCD Out Size            */
		writel(imc_reserve->l1a->size, dev->IMCMmioV + IMC_L1A_SIZE);
	}

	if (imc_reserve->l1b) {
		/*** Layer1b Setting Register ***/
		/* Layer1B Control                 */
		writel(imc_reserve->l1b->control,
		 dev->IMCMmioV + IMC_L1B_CONTROL);
		/* Layer1B Color Key Control       */
		writel(imc_reserve->l1b->keyenable,
		 dev->IMCMmioV + IMC_L1B_KEYENABLE);
		/* Layer1B Alpha                   */
		writel(imc_reserve->l1b->alpha, dev->IMCMmioV + IMC_L1B_ALPHA);
		/* Layer1B Address                 */
		writel(imc_reserve->l1b->frameadr,
		 dev->IMCMmioV + IMC_L1B_FRAMEADR);
		/* Layer1B LCD Out Positon         */
		writel(imc_reserve->l1b->position,
		 dev->IMCMmioV + IMC_L1B_POSITION);
		/* Layer1B LCD Out Size            */
		writel(imc_reserve->l1b->size, dev->IMCMmioV + IMC_L1B_SIZE);
	}

	if (imc_reserve->l1c) {
		/*** Layer1c Setting Register ***/
		/* Layer1C Control                 */
		writel(imc_reserve->l1c->control,
		 dev->IMCMmioV + IMC_L1C_CONTROL);
		/* Layer1C Color Key Control       */
		writel(imc_reserve->l1c->keyenable,
		 dev->IMCMmioV + IMC_L1C_KEYENABLE);
		/* Layer1C Alpha                   */
		writel(imc_reserve->l1c->alpha, dev->IMCMmioV + IMC_L1C_ALPHA);
		/* Layer1C Address                 */
		writel(imc_reserve->l1c->frameadr,
		 dev->IMCMmioV + IMC_L1C_FRAMEADR);
		/* Layer1C LCD Out Positon         */
		writel(imc_reserve->l1c->position,
		 dev->IMCMmioV + IMC_L1C_POSITION);
		/* Layer1C LCD Out Size            */
		writel(imc_reserve->l1c->size, dev->IMCMmioV + IMC_L1C_SIZE);
	}

	if (imc_reserve->l2a) {
		/*** Layer2a Setting Register ***/
		/* Layer2A Control                 */
		writel(imc_reserve->l2a->control,
		 dev->IMCMmioV + IMC_L2A_CONTROL);
		/* Layer2A Format                  */
		writel(imc_reserve->l2a->format,
		 dev->IMCMmioV + IMC_L2A_FORMAT);
		/* Layer2A Double Buffer Control   */
		writel(imc_reserve->l2a->bufsel,
		 dev->IMCMmioV + IMC_L2A_BUFSEL);
		/* Layer2A Resize                  */
		writel(imc_reserve->l2a->resize,
		 dev->IMCMmioV + IMC_L2A_RESIZE);
		/* Layer2A Horizontal, Vertical Mirroring Control */
		writel(imc_reserve->l2a->mirror,
		 dev->IMCMmioV + IMC_L2A_MIRROR);
		/* Layer2A Address increment value */
		writel(imc_reserve->l2a->offset,
		 dev->IMCMmioV + IMC_L2A_OFFSET);
		/* Layer2A Address (YP)            */
		writel(imc_reserve->l2a->frameadr_yp,
		 dev->IMCMmioV + IMC_L2A_FRAMEADR_YP);
		/* Layer2A Address (UP)            */
		writel(imc_reserve->l2a->frameadr_up,
		 dev->IMCMmioV + IMC_L2A_FRAMEADR_UP);
		/* Layer2A Address (VP)            */
		writel(imc_reserve->l2a->frameadr_vp,
		 dev->IMCMmioV + IMC_L2A_FRAMEADR_VP);
		/* Layer2A Address (YQ)            */
		writel(imc_reserve->l2a->frameadr_yq,
		 dev->IMCMmioV + IMC_L2A_FRAMEADR_YQ);
		/* Layer2A Address (UQ)            */
		writel(imc_reserve->l2a->frameadr_uq,
		 dev->IMCMmioV + IMC_L2A_FRAMEADR_UQ);
		/* Layer2A Address (VQ)            */
		writel(imc_reserve->l2a->frameadr_vq,
		 dev->IMCMmioV + IMC_L2A_FRAMEADR_VQ);
		/* Layer2A LCD Out Positon         */
		writel(imc_reserve->l2a->position,
		 dev->IMCMmioV + IMC_L2A_POSITION);
		/* Layer2A LCD Out Size            */
		writel(imc_reserve->l2a->size, dev->IMCMmioV + IMC_L2A_SIZE);
	}

	if (imc_reserve->l2b) {
		/*** Layer2b Setting Register ***/
		/* Layer2B Control                 */
		writel(imc_reserve->l2b->control,
		 dev->IMCMmioV + IMC_L2B_CONTROL);
		/* Layer2B Format                  */
		writel(imc_reserve->l2b->format,
		 dev->IMCMmioV + IMC_L2B_FORMAT);
		/* Layer2B Double Buffer Control   */
		writel(imc_reserve->l2b->bufsel,
		 dev->IMCMmioV + IMC_L2B_BUFSEL);
		/* Layer2B Resize                  */
		writel(imc_reserve->l2b->resize,
		 dev->IMCMmioV + IMC_L2B_RESIZE);
		/* Layer2B Horizontal, Vertical Mirroring Control */
		writel(imc_reserve->l2b->mirror,
		 dev->IMCMmioV + IMC_L2B_MIRROR);
		/* Layer2B Address increment value */
		writel(imc_reserve->l2b->offset,
		 dev->IMCMmioV + IMC_L2B_OFFSET);
		/* Layer2B Address (YP)            */
		writel(imc_reserve->l2b->frameadr_yp,
		 dev->IMCMmioV + IMC_L2B_FRAMEADR_YP);
		/* Layer2B Address (UP)            */
		writel(imc_reserve->l2b->frameadr_up,
		 dev->IMCMmioV + IMC_L2B_FRAMEADR_UP);
		/* Layer2B Address (VP)            */
		writel(imc_reserve->l2b->frameadr_vp,
		 dev->IMCMmioV + IMC_L2B_FRAMEADR_VP);
		/* Layer2B Address (YQ)            */
		writel(imc_reserve->l2b->frameadr_yq,
		 dev->IMCMmioV + IMC_L2B_FRAMEADR_YQ);
		/* Layer2B Address (UQ)            */
		writel(imc_reserve->l2b->frameadr_uq,
		 dev->IMCMmioV + IMC_L2B_FRAMEADR_UQ);
		/* Layer2B Address (VQ)            */
		writel(imc_reserve->l2b->frameadr_vq,
		 dev->IMCMmioV + IMC_L2B_FRAMEADR_VQ);
		/* Layer2B LCD Out Positon         */
		writel(imc_reserve->l2b->position,
		 dev->IMCMmioV + IMC_L2B_POSITION);
		/* Layer2B LCD Out Size            */
		writel(imc_reserve->l2b->size, dev->IMCMmioV + IMC_L2B_SIZE);
	}

	if (imc_reserve->bg) {
		/*** LayerBG Setting Register ***/
		/* LayerBG Format                  */
		writel(imc_reserve->bg->format, dev->IMCMmioV + IMC_BG_FORMAT);
		/* LayerBG Resize                  */
		writel(imc_reserve->bg->resize, dev->IMCMmioV + IMC_BG_RESIZE);
		/* LayerBG Address increment value */
		writel(imc_reserve->bg->offset, dev->IMCMmioV + IMC_BG_OFFSET);
		/* LayerBG Address                 */
		writel(imc_reserve->bg->frameadr,
		 dev->IMCMmioV + IMC_BG_FRAMEADR);
	}

	dev->callback_func_refresh = callback_refresh;

	imc_hw_int_enable();

	/* register REFRESH */
	writel(IMC_UPDATE_ON,	dev->IMCMmioV + IMC_REFRESH);

	/*<<<<<< interuppt enable */
	spin_unlock_irqrestore(&dev->imc_hw_lock, flags);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
	return 0;
}


#if LCD_ASYNCHRONOUS /* this function checked to IMC-LCD asynchronous mode. */
/******************************************************************************
* MODULE   : mp200_imc_start
* FUNCTION : this function used in IMC-LCD asynchronous mode.
* RETURN   :   0: success
*          : -13: EACCES
*          : -16: EBUSY
* NOTE     : none
* CREATE   : 2007.06.21 M.Haguro
* UPDATE   :
******************************************************************************/
int  mp200_imc_start(void)
{
	unsigned long flags;

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	if (!imc_hw_chk_startmode()) {
		printk_err("error! IMC_CONTROL is \"LCD SYNCRO MODE\"\n");
		return -EACCES;
	}

	if (!dev->chk_preset) {
		printk_err("error! need to \"mp200_imc_preset()\"\n");
		return -EACCES;
	}

	if (imc_hw_chk_busy()) {
		/* IMC running */
		printk_err("IMC WORKING NOW\n");
		return -EBUSY;
	}

	/*>>>>>> interuppt disable */
	spin_lock_irqsave(&dev->imc_hw_lock, flags);

	imc_hw_int_enable();

	/* Start to IMC */
	writel(IMC_IMCSTART_ON, dev->IMCMmioV + IMC_START);

	/*<<<<<< interuppt enable */
	spin_unlock_irqrestore(&dev->imc_hw_lock, flags);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
	return 0;
}
#endif


/******************************************************************************
* MODULE   : mp200_imc_set_callback
* FUNCTION :
* RETURN   :   0: success
*          : -22: EINVAL : callback is invalid adrress
* NOTE     : none
* CREATE   : 2007.06.21 M.Haguro
* UPDATE   :
******************************************************************************/
int mp200_imc_set_callback(imc_callback_func_refresh callback_refresh,
 imc_callback_func_wb callback_wb)
{
	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	if (!callback_refresh) {
		printk_err("error! imc_callback_func_refresh is invalid address\n");
		return -EINVAL;
	}
	if (!callback_wb) {
		printk_err("error! imc_callback_func_wb is invalid address\n");
		return -EINVAL;
	}

	dev->callback_func_refresh = callback_refresh;
	dev->callback_func_wb      = callback_wb;
	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
	return 0;
}


/* ----- internal function ------------------------------------------------- */
/******************************************************************************
* MODULE   : imc_hw_init
* FUNCTION :
* RETURN   :   0: success
*          : -16: EBUSY
* NOTE     : none
* CREATE   : 2007.06.15 M.Haguro
* UPDATE   :
******************************************************************************/
static int imc_hw_init(void)
{
	unsigned long flags;

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	if (imc_hw_chk_busy()) {
		/* IMC running */
		printk_err("IMC WORKING NOW\n");
		return -EBUSY;
	}

	/*>>>>>> interuppt disable */
	spin_lock_irqsave(&dev->imc_hw_lock, flags);

	imc_hw_unreset();

	imc_hw_int_disable();

	/*<<<<<< interuppt enable */
	spin_unlock_irqrestore(&dev->imc_hw_lock, flags);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
	return 0;
}


/******************************************************************************
* MODULE   : imc_hw_reset
* FUNCTION : reset IMC module
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.06.18 M.Haguro
* UPDATE   :
******************************************************************************/
void imc_hw_reset(void)
{
	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	/* Reset IMC */
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_IMC);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_IMCPCLK);
	mp200_pmu_reset_device(MP200_RESETDEVICE_IMC);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_IMC_CLK);
	mp200_pmu_close_clockgate(MP200_CLOCKGATE_IMC_PCLK);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
}


/******************************************************************************
* MODULE   : imc_hw_unreset
* FUNCTION : un-reset IMC module
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.06.18 M.Haguro
* UPDATE   :
******************************************************************************/
void imc_hw_unreset(void)
{
	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	/* UnReset IMC */
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_IMC_PCLK);
	mp200_pmu_open_clockgate(MP200_CLOCKGATE_IMC_CLK);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_IMC);
	mp200_pmu_clkctrl_off(MP200_CLKCTRL_IMCPCLK);
	mp200_pmu_unreset_device(MP200_RESETDEVICE_IMC);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_IMC);
	mp200_pmu_clkctrl_on(MP200_CLKCTRL_IMCPCLK);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
}


/******************************************************************************
* MODULE   : imc_hw_int_enable
* FUNCTION :
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.06.20 M.Haguro
* UPDATE   :
******************************************************************************/
static void imc_hw_int_enable(void)
{

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	/* Interuppt Control Register */
	/* Interuppt Enable Set   */
	writel(IMC_INTENSET_ALL_BIT, dev->IMCMmioV + IMC_INTENSET);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
}


/******************************************************************************
* MODULE   : imc_hw_int_disable
* FUNCTION :
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.06.20 M.Haguro
* UPDATE   :
******************************************************************************/
static void imc_hw_int_disable(void)
{
	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	/* Interuppt Control Register */
	/* Interuppt Enable Clear */
	writel(IMC_INTENCLR_ALL_BIT, dev->IMCMmioV + IMC_INTENCLR);
	/* Interuppt FF Clear     */
	writel(IMC_INTFFCLR_ALL_BIT, dev->IMCMmioV + IMC_INTFFCLR);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
}


#if LCD_ASYNCHRONOUS /* this function checked to IMC-LCD asynchronous mode. */
/******************************************************************************
* MODULE   : imc_hw_chk_startmode
* FUNCTION :
* RETURN   : 0: LCD Syncro
*          : 1: No Sync
* NOTE     : none
* CREATE   : 2007.06.22 M.Haguro
* UPDATE   :
******************************************************************************/
static int imc_hw_chk_startmode(void)
{
	int iRet;
	/* Control for IMC */
	unsigned long ulRegVal32 = readl(dev->IMCMmioV + IMC_CONTROL);

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");
	iRet = (((ulRegVal32 & IMC_START_MODE_BIT) == IMC_START_MODE_NOSYNC) ? 1 : 0);
	printk_dbg((_DEBUG_IMC & 0x02), "out\n");

	return iRet;
}
#endif


/******************************************************************************
* MODULE   : imc_hw_chk_busy
* FUNCTION :
* RETURN   :  0: IMC stop
*          : -1: IMC busy
* NOTE     : none
* CREATE   : 2007.06.22 M.Haguro
* UPDATE   :
******************************************************************************/
static int imc_hw_chk_busy(void)
{
	int iRet;
	/* Status for IMC */
	unsigned long ulRegVal32 = readl(dev->IMCMmioV + IMC_STATUS);

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");
	iRet = (((ulRegVal32 & IMC_STATUS_BIT) == IMC_STATUS_STOP) ? 0 : -1);
	printk_dbg((_DEBUG_IMC & 0x02), "out\n");

	return iRet;
}


#if defined(CONFIG_PM) || defined(CONFIG_DPM)
/******************************************************************************
* MODULE   : imc_hw_chk_status
* FUNCTION : enabled suspend
* RETURN   :  0 : success
*            -1 : failed
* NOTE     : none
* CREATE   : 2007.11.19 M.Haguro
* UPDATE   :
******************************************************************************/
static int imc_hw_chk_status()
{
	int iRet, i = 0;

	do {
		iRet = imc_hw_chk_busy();
		if (iRet == 0) {
			i = 5;
		} else if (i == 4) {
			printk_err("imc_hw_chk_status\n");
			Call_StatusCtrlFunc_OFF();
			return -1;
		} else {
			Call_StatusCtrlFunc_ON();
			mdelay(16);
			i++;
		}
	} while (i < 5);

	return 0;
}
#endif /* CONFIG_PM || CONFIG_DPM */


/* ----- IMC IRQ handler --------------------------------------------------- */
/******************************************************************************
* MODULE   : imc_irq_handler
* FUNCTION :
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.06.22 M.Haguro
* UPDATE   :
******************************************************************************/
static irqreturn_t imc_irq_handler(int irq, void *dev_id)
{
	imc_irq_handler_sub(irq, dev_id);
	return IRQ_HANDLED;
}


/******************************************************************************
* MODULE   : imc_irq_handler_sub
* FUNCTION :
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.07.06 M.Haguro
* UPDATE   :
******************************************************************************/
static void imc_irq_handler_sub(int irq, void *dev_id)
{
	unsigned long ulRegVal32 = readl(dev->IMCMmioV + IMC_INTSTATUS);

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	if (ulRegVal32 & (IMC_REFRESH_BIT | IMC_WBEND_BIT)) {
		if (ulRegVal32 & IMC_REFRESH_BIT) {
			printk_dbg((_DEBUG_IMC & 0x02), "IMC_REFRESH\n");
			/* Interuppt FF Clear     */
			writel(IMC_REFRESH_CLR_BIT,
			 dev->IMCMmioV + IMC_INTFFCLR);
			if (dev->callback_func_refresh)
				(dev->callback_func_refresh)();
#if LCD_ASYNCHRONOUS /* this function checked to IMC-LCD asynchronous mode. */
			if (dev->callback_func_wb)
				(dev->callback_func_wb)(ulRegVal32);
#else
		}
		if (ulRegVal32 & IMC_WBEND_BIT) {
			printk_dbg((_DEBUG_IMC & 0x02), "IMC_WBEND\n");
			/* Interuppt FF Clear     */
			writel(IMC_WBEND_CLR_BIT, dev->IMCMmioV + IMC_INTFFCLR);
			if (dev->callback_func_wb)
				(dev->callback_func_wb)(ulRegVal32);
#endif
		}
	}

	/* error interrupt */
	else if (ulRegVal32 & IMC_AHBRERR_BIT) {
		printk_err("IMC AHBR ERROR. Reset IMC.\n");
		/* Interuppt FF Clear     */
		writel(IMC_AHBRERR_CLR_BIT,	dev->IMCMmioV + IMC_INTFFCLR);
		imc_hw_reset();
		imc_hw_unreset();
		dev->chk_preset = 0;
		if (dev->callback_func_wb)
			(dev->callback_func_wb)(ulRegVal32);

	} else if (ulRegVal32 & IMC_AHBWERR_BIT) {
		printk_err("IMC AHBW ERROR. Reset IMC.\n");
		/* Interuppt FF Clear     */
		writel(IMC_AHBWERR_CLR_BIT,	dev->IMCMmioV + IMC_INTFFCLR);
		imc_hw_reset();
		imc_hw_unreset();
		dev->chk_preset = 0;
		if (dev->callback_func_wb)
			(dev->callback_func_wb)(ulRegVal32);

	} else if (ulRegVal32 & IMC_OVERRUN_BIT) {
		printk_err("IMC OVERRUN ERROR. Retry IMC.\n");
		/* Interuppt FF Clear     */
		writel(IMC_OVERRUN_CLR_BIT,	dev->IMCMmioV + IMC_INTFFCLR);
	}

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
}




#if defined(CONFIG_PM) || defined(CONFIG_DPM)
/* ----- suspend/resume function ------------------------------------------- */
/******************************************************************************
* MODULE   : mp200_imc_suspend
* FUNCTION : suspend IMC driver
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.11.06 M.haguro
* UPDATE   :
******************************************************************************/
int mp200_imc_suspend(struct platform_device *dev, pm_message_t state)
{
	if (state.event == DEV_SUSPEND_IDLE_1) {
		if (imc_hw_chk_busy())
			return -EBUSY;
		else
			return 0;
	} else {/* state.event == PM_EVENT_SUSPEND */

		if (mp200_sleep_while_idle) {
			return 0;
		} else {
			if (imc_hw_chk_status())
				return -1;

			DPM_suspend_flg = 1;
			imc_hw_reset(); /* Reset IMC */
			return 0;
		}
	}
}


/******************************************************************************
* MODULE   : mp200_imc_resume
* FUNCTION : suspend IMC driver
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.11.06 M.haguro
* UPDATE   :
******************************************************************************/
int mp200_imc_resume(struct platform_device *dev)
{
	if (DPM_suspend_flg) {
		DPM_suspend_flg = 0;
		imc_hw_unreset(); /* UnReset IMC */

		if (ctrl_func) {
			/* start the next request */
			tasklet_schedule(&imc_nexttask);
		}
	}
	return 0;
}


/******************************************************************************
* MODULE   : imc_resume_kick
* FUNCTION : suspend IMC driver. call TASKLET
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.11.06 M.haguro
* UPDATE   :
******************************************************************************/
static void imc_resume_kick(unsigned long num)
{
	Call_StatusCtrlFunc_OFF();

	if (dev->callback_func_wb) {
		printk_dbg((_DEBUG_IMC & 0x02), "\n");
		(dev->callback_func_wb)(IMC_WBEND_BIT);
	}
}
#endif /* CONFIG_PM || CONFIG_DPM */


/* ----- init function ----------------------------------------------------- */
/******************************************************************************
* MODULE   : mp200_imc_init_module
* FUNCTION :
* RETURN   : 0    : success
*            other: fail
* NOTE     : none
* CREATE   : 2007.06.15 M.Haguro
* UPDATE   :
******************************************************************************/
int mp200_imc_init_module(void)
{
	int iRet = 0;

	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	/* allocate device infomation */
	dev = kmalloc(sizeof(struct mp200_imc_device), GFP_KERNEL);
	if (!dev) {
		printk_err("could not allocate memory\n");
		iRet = -ENOMEM;
		goto alloc_err;
	}
	memset(dev, 0, sizeof(struct mp200_imc_device));
	dev->IMCMmio       = IMC_BASE;
	dev->IMCMmioLength = IMC_SIZE;

	/* ioremap IMCMmio */
#ifdef IO_ADDRESS
	dev->IMCMmioV = (char *)IO_ADDRESS(dev->IMCMmio);
#else
	dev->IMCMmioV = ioremap_nocache(dev->IMCMmio, dev->IMCMmioLength);
	if (!dev->IMCMmioV) {
		printk_err("cannot ioremap_nocache IMCMmio\n");
		iRet = -ENOMEM;
		goto ioremap_err;
	}
#endif
	printk_dbg((_DEBUG_IMC),
	 "IMCMmio (0x%lx)  IMCMmioV (0x%p)  IMCMmioLength (0x%lx)\n",
	  dev->IMCMmio, dev->IMCMmioV, dev->IMCMmioLength);

	/* IMC initialize */
	iRet = imc_hw_init();
	if (iRet)
		goto init_hw_err;

	/* spinlock initialize */
	spin_lock_init(&dev->imc_hw_lock);

	/* install IMC irq handler */
	iRet = request_irq(INT_IMC, imc_irq_handler, 0, DEV_NAME, NULL);
	if (iRet) {
		printk_err("fail in request_irq(INT_IMC)\n");
		goto init_hw_err;
	}

	goto init_success;

init_hw_err:
#ifndef IO_ADDRESS
	iounmap(dev->IMCMmioV);
ioremap_err:
#endif
	kfree(dev);
alloc_err:
init_success:

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
	return iRet;
}


#ifdef MODULE
/******************************************************************************
* MODULE   : mp200_imc_exit_module
* FUNCTION :
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.06.18 M.Haguro
* UPDATE   :
******************************************************************************/
void mp200_imc_exit_module(void)
{
	printk_dbg((_DEBUG_IMC & 0x01), "in\n");

	imc_hw_int_disable();
	free_irq(INT_IMC, NULL);
	imc_hw_reset();
#ifndef IO_ADDRESS
	iounmap(dev->IMCMmioV);
#endif
	kfree(dev);

	printk_dbg((_DEBUG_IMC & 0x02), "out\n");
}


module_init(mp200_imc_init_module);
module_exit(mp200_imc_exit_module);
#else
device_initcall(mp200_imc_init_module);
#endif


MODULE_AUTHOR("NEC Electronics Corporation.");
MODULE_DESCRIPTION("EM1 IMC control driver");
MODULE_LICENSE("GPL");

